"""Ledger configuration."""

import logging
import re
import sys
from collections import OrderedDict
from typing import Optional

import markdown
import prompt_toolkit
from prompt_toolkit.formatted_text import HTML
from uuid_utils import uuid4

from ..config.settings import Settings
from ..core.profile import Profile
from ..ledger.base import BaseLedger
from ..ledger.endpoint_type import EndpointType
from ..ledger.error import LedgerError
from ..utils.http import FetchError, fetch
from ..wallet.base import BaseWallet
from .base import ConfigError

LOGGER = logging.getLogger(__name__)


async def fetch_genesis_transactions(genesis_url: str) -> str:
    """Get genesis transactions."""
    headers = {}
    headers["Content-Type"] = "application/json"
    LOGGER.info("Fetching genesis transactions from: %s", genesis_url)
    try:
        # Fetch from --genesis-url likely to fail in composed container setup
        # https://github.com/openwallet-foundation/acapy/issues/1745
        return await fetch(genesis_url, headers=headers, max_attempts=20)
    except FetchError as e:
        LOGGER.error("Error retrieving genesis transactions from %s: %s", genesis_url, e)
        raise ConfigError("Error retrieving ledger genesis transactions") from e


async def fetch_genesis_from_url_or_file(
    genesis_url: Optional[str], genesis_path: Optional[str]
) -> str:
    """Fetch genesis transactions from URL or file."""
    txns = ""
    if genesis_url:
        txns = await fetch_genesis_transactions(genesis_url)
    elif genesis_path:
        try:
            LOGGER.info("Reading ledger genesis transactions from: %s", genesis_path)
            with open(genesis_path, "r") as genesis_file:
                txns = genesis_file.read()
        except IOError as e:
            LOGGER.error("Failed to read genesis file: %s", str(e))
            raise ConfigError("Error reading ledger genesis transactions") from e
    else:
        LOGGER.warning("No genesis url or path found in settings")
    return txns


async def get_genesis_transactions(settings: Settings) -> str:
    """Fetch genesis transactions if necessary."""

    LOGGER.debug("Getting genesis transactions from settings")
    txns = settings.get("ledger.genesis_transactions")
    LOGGER.debug("Genesis transactions from settings: %s", "found" if txns else "absent")
    if not txns:
        LOGGER.debug("No genesis transactions found in settings")
        genesis_url = settings.get("ledger.genesis_url")
        genesis_path = settings.get("ledger.genesis_file")

        txns = await fetch_genesis_from_url_or_file(genesis_url, genesis_path)
        if txns:
            LOGGER.debug("Storing genesis transactions in settings")
            settings["ledger.genesis_transactions"] = txns

    return txns


async def load_multiple_genesis_transactions_from_config(settings: Settings) -> None:
    """Fetch genesis transactions for multiple ledger configuration."""

    ledger_config_list = settings.get("ledger.ledger_config_list")
    ledger_txns_list = []
    write_ledger_set = False

    LOGGER.debug("Processing %d ledger configs", len(ledger_config_list))
    for config in ledger_config_list:
        txns = config.get("genesis_transactions")

        if not txns:
            genesis_url = config.get("genesis_url")
            genesis_path = config.get("genesis_file")
            txns = await fetch_genesis_from_url_or_file(genesis_url, genesis_path)

        is_write_ledger = config.get("is_write", False)
        if is_write_ledger:
            write_ledger_set = True

        ledger_id = config.get("id", str(uuid4()))  # Default to UUID if no ID provided
        config_item = {
            "id": ledger_id,
            "is_production": config.get("is_production", True),
            "is_write": is_write_ledger,
            "genesis_transactions": txns,
            "keepalive": int(config.get("keepalive", 5)),
            "read_only": bool(config.get("read_only", False)),
            "socks_proxy": config.get("socks_proxy"),
            "pool_name": config.get("pool_name", ledger_id),
        }
        if "endorser_alias" in config:
            config_item["endorser_alias"] = config.get("endorser_alias")
        if "endorser_did" in config:
            config_item["endorser_did"] = config.get("endorser_did")

        ledger_txns_list.append(config_item)

    # Check if we have a writable ledger or genesis information
    is_read_only = settings.get("ledger.read_only")
    has_genesis_info = (
        settings.get("ledger.genesis_transactions")
        or settings.get("ledger.genesis_file")
        or settings.get("ledger.genesis_url")
    )

    # Raise error if we have neither a writable ledger nor genesis info (unless read-only)
    if not write_ledger_set and not is_read_only and not has_genesis_info:
        raise ConfigError(
            "No writable ledger configured and no genesis information provided. "
            "Please set is_write=True for a ledger or provide genesis_url, "
            "genesis_file, or genesis_transactions."
        )

    settings["ledger.ledger_config_list"] = ledger_txns_list
    LOGGER.debug("Processed %d ledger configs successfully", len(ledger_txns_list))


async def ledger_config(
    profile: Profile, public_did: str, provision: bool = False
) -> bool:
    """Perform Indy ledger configuration."""

    LOGGER.debug(
        "Configuring ledger for profile %s and public_did %s", profile.name, public_did
    )

    session = await profile.session()

    ledger = session.inject_or(BaseLedger)
    if not ledger:
        LOGGER.info("Ledger instance not provided")
        return False

    async with ledger:
        # Check transaction author agreement acceptance
        if not ledger.read_only:
            LOGGER.debug("Checking transaction author agreement")
            taa_info = await ledger.get_txn_author_agreement()
            if taa_info["taa_required"] and public_did:
                LOGGER.debug("TAA acceptance required")
                taa_accepted = await ledger.get_latest_txn_author_acceptance()

                taa_record_digest = taa_info["taa_record"]["digest"]  # keys exist
                taa_accepted_digest = taa_accepted.get("digest")  # key might not exist

                digest_match = taa_record_digest == taa_accepted_digest
                if not taa_accepted or not digest_match:
                    LOGGER.info("TAA acceptance needed - performing acceptance")
                    if not await accept_taa(ledger, profile, taa_info, provision):
                        LOGGER.warning("TAA acceptance failed")
                        return False
                    LOGGER.info("TAA acceptance completed")

        # Publish endpoints if necessary - skipped if TAA is required but not accepted
        endpoint = session.settings.get("default_endpoint")
        if public_did:
            wallet = session.inject(BaseWallet)
            try:
                LOGGER.debug("Setting DID endpoint to: %s", endpoint)
                await wallet.set_did_endpoint(public_did, endpoint, ledger)
            except LedgerError as x_ledger:
                LOGGER.error("Error setting DID endpoint: %s", x_ledger.message)
                raise ConfigError(x_ledger.message) from x_ledger  # e.g., read-only

            # Publish profile endpoint if ledger is NOT read-only
            profile_endpoint = session.settings.get("profile_endpoint")
            if profile_endpoint and not ledger.read_only:
                LOGGER.debug(
                    "Publishing profile endpoint: %s for DID: %s",
                    profile_endpoint,
                    public_did,
                )
                await ledger.update_endpoint_for_did(
                    public_did, profile_endpoint, EndpointType.PROFILE
                )
                LOGGER.info("Profile endpoint published successfully")

    LOGGER.info("Ledger configuration complete")
    return True


async def select_aml_tty(taa_info, provision: bool = False) -> Optional[str]:
    """Select acceptance mechanism from AML."""
    mechanisms = taa_info["aml_record"]["aml"]
    allow_opts = OrderedDict(
        [
            (
                "wallet_agreement",
                (
                    "Accept the transaction author agreement and store the "
                    "acceptance in the wallet"
                ),
            ),
            (
                "on_file",
                (
                    "Acceptance of the transaction author agreement is on file "
                    "in my organization"
                ),
            ),
        ]
    )
    if not provision:
        allow_opts["for_session"] = (
            "Accept the transaction author agreement for the duration of "
            "the current session"
        )

    found = []
    for opt in allow_opts:
        if opt in mechanisms:
            found.append(opt)

    md = markdown.Markdown()
    taa_html = md.convert(taa_info["taa_record"]["text"])
    taa_html = re.sub(
        r"<h[1-6]>(.*?)</h[1-6]>", r"<p><strong>\1</strong></p>\n", taa_html
    )
    taa_html = re.sub(r"<li>(.*?)</li>", r" - \1", taa_html)

    taa_html = (
        "\n<strong>Transaction Author Agreement version "
        + taa_info["taa_record"]["version"]
        + "</strong>\n\n"
        + taa_html
    )

    prompt_toolkit.print_formatted_text(HTML(taa_html))

    opts = []
    num_mechanisms = {}
    for idx, opt in enumerate(found):
        num_mechanisms[str(idx + 1)] = opt
        opts.append(f" {idx + 1}. {allow_opts[opt]}")
    opts.append(" X. Skip the transaction author agreement")
    opts_text = "\nPlease select an option:\n" + "\n".join(opts) + "\n[1]> "

    while True:
        try:
            opt = await prompt_toolkit.prompt(opts_text, async_=True)
        except EOFError:
            return None
        if not opt:
            opt = "1"
        opt = opt.strip()
        if opt in ("x", "X"):
            return None
        if opt in num_mechanisms:
            mechanism = num_mechanisms[opt]
            break

    return mechanism


async def accept_taa(
    ledger: BaseLedger,
    profile: Profile,
    taa_info,
    provision: bool = False,
) -> bool:
    """Perform TAA acceptance."""

    mechanisms = taa_info["aml_record"]["aml"]
    mechanism = None

    taa_acceptance_mechanism = profile.settings.get("ledger.taa_acceptance_mechanism")
    taa_acceptance_version = profile.settings.get("ledger.taa_acceptance_version")

    # If configured, accept the TAA automatically
    if taa_acceptance_mechanism:
        taa_record_version = taa_info["taa_record"]["version"]
        if taa_acceptance_version != taa_record_version:
            raise LedgerError(
                f"TAA version ({taa_record_version}) is different from TAA accept "
                f"version ({taa_acceptance_version}) from configuration. Update the "
                "TAA version in the config to accept the TAA."
            )

        if taa_acceptance_mechanism not in mechanisms:
            valid_mechanisms = ", ".join(mechanisms.keys())
            raise LedgerError(
                f"TAA acceptance mechanism '{taa_acceptance_mechanism}' is not a "
                f"valid acceptance mechanism. Valid mechanisms are: {valid_mechanisms}"
            )

        mechanism = taa_acceptance_mechanism
    # If tty is available use it (allows to accept newer TAA than configured)
    elif sys.stdout.isatty():
        mechanism = await select_aml_tty(taa_info, provision)
    else:
        LOGGER.warning(
            "Cannot accept TAA without interactive terminal or taa accept config"
        )

    if not mechanism:
        return False

    LOGGER.debug(f"Accepting the TAA using mechanism '{mechanism}'")
    await ledger.accept_txn_author_agreement(taa_info["taa_record"], mechanism)
    return True



"""Provision command for setting up agent settings before starting."""

import asyncio
import logging
from typing import Sequence

from configargparse import ArgumentParser

from ..config import argparse as arg
from ..config.base import BaseError
from ..config.default_context import DefaultContextBuilder
from ..config.ledger import (
    get_genesis_transactions,
    ledger_config,
    load_multiple_genesis_transactions_from_config,
)
from ..config.util import common_config
from ..config.wallet import wallet_config
from ..protocols.coordinate_mediation.mediation_invite_store import (
    MediationInviteRecord,
    MediationInviteStore,
)
from ..storage.base import BaseStorage
from . import PROG

LOGGER = logging.getLogger(__name__)


class ProvisionError(BaseError):
    """Base exception for provisioning errors."""


def init_argument_parser(parser: ArgumentParser):
    """Initialize an argument parser with the module's arguments."""
    return arg.load_argument_groups(parser, *arg.group.get_registered(arg.CAT_PROVISION))


async def provision(settings: dict):
    """Perform provisioning."""
    context_builder = DefaultContextBuilder(settings)
    context = await context_builder.build_context()

    try:
        if context.settings.get("ledger.ledger_config_list"):
            await load_multiple_genesis_transactions_from_config(context.settings)
        if (
            context.settings.get("ledger.genesis_transactions")
            or context.settings.get("ledger.genesis_file")
            or context.settings.get("ledger.genesis_url")
        ):
            await get_genesis_transactions(context.settings)

        root_profile, public_did = await wallet_config(context, provision=True)

        # store mediator invite url if provided
        mediation_invite = settings.get("mediation.invite", None)
        if mediation_invite:
            async with root_profile.session() as session:
                await MediationInviteStore(session.context.inject(BaseStorage)).store(
                    MediationInviteRecord.unused(mediation_invite)
                )

        if await ledger_config(root_profile, public_did and public_did.did, True):
            LOGGER.info("Ledger configured")
        else:
            LOGGER.warning("Ledger not configured")

        await root_profile.close()
    except BaseError as e:
        raise ProvisionError("Error during provisioning") from e


def execute(argv: Sequence[str] = None):
    """Entrypoint."""
    parser = arg.create_argument_parser(prog=PROG)
    parser.prog += " provision"
    get_settings = init_argument_parser(parser)
    args = parser.parse_args(argv)
    settings = get_settings(args)
    common_config(settings)

    # provision needs write access to the ledger
    # (override if specified otherwise)
    settings["ledger.read_only"] = False

    loop = asyncio.get_event_loop()
    loop.run_until_complete(provision(settings))


def main():
    """Execute the main line."""
    if __name__ == "__main__":
        execute()


main()



CREDENTIALS_V2 = {
    "@context": {
        "@protected": True,
        "id": "@id",
        "type": "@type",
        "description": "https://schema.org/description",
        "digestMultibase": {
            "@id": "https://w3id.org/security#digestMultibase",
            "@type": "https://w3id.org/security#multibase",
        },
        "digestSRI": {
            "@id": "https://www.w3.org/2018/credentials#digestSRI",
            "@type": "https://www.w3.org/2018/credentials#sriString",
        },
        "mediaType": {"@id": "https://schema.org/encodingFormat"},
        "name": "https://schema.org/name",
        "VerifiableCredential": {
            "@id": "https://www.w3.org/2018/credentials#VerifiableCredential",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "confidenceMethod": {
                    "@id": "https://www.w3.org/2018/credentials#confidenceMethod",
                    "@type": "@id",
                },
                "credentialSchema": {
                    "@id": "https://www.w3.org/2018/credentials#credentialSchema",
                    "@type": "@id",
                },
                "credentialStatus": {
                    "@id": "https://www.w3.org/2018/credentials#credentialStatus",
                    "@type": "@id",
                },
                "credentialSubject": {
                    "@id": "https://www.w3.org/2018/credentials#credentialSubject",
                    "@type": "@id",
                },
                "description": "https://schema.org/description",
                "evidence": {
                    "@id": "https://www.w3.org/2018/credentials#evidence",
                    "@type": "@id",
                },
                "issuer": {
                    "@id": "https://www.w3.org/2018/credentials#issuer",
                    "@type": "@id",
                },
                "name": "https://schema.org/name",
                "proof": {
                    "@id": "https://w3id.org/security#proof",
                    "@type": "@id",
                    "@container": "@graph",
                },
                "refreshService": {
                    "@id": "https://www.w3.org/2018/credentials#refreshService",
                    "@type": "@id",
                },
                "relatedResource": {
                    "@id": "https://www.w3.org/2018/credentials#relatedResource",
                    "@type": "@id",
                },
                "renderMethod": {
                    "@id": "https://www.w3.org/2018/credentials#renderMethod",
                    "@type": "@id",
                },
                "termsOfUse": {
                    "@id": "https://www.w3.org/2018/credentials#termsOfUse",
                    "@type": "@id",
                },
                "validFrom": {
                    "@id": "https://www.w3.org/2018/credentials#validFrom",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "validUntil": {
                    "@id": "https://www.w3.org/2018/credentials#validUntil",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
            },
        },
        "EnvelopedVerifiableCredential": "https://www.w3.org/2018/credentials#EnvelopedVerifiableCredential",
        "VerifiablePresentation": {
            "@id": "https://www.w3.org/2018/credentials#VerifiablePresentation",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "holder": {
                    "@id": "https://www.w3.org/2018/credentials#holder",
                    "@type": "@id",
                },
                "proof": {
                    "@id": "https://w3id.org/security#proof",
                    "@type": "@id",
                    "@container": "@graph",
                },
                "termsOfUse": {
                    "@id": "https://www.w3.org/2018/credentials#termsOfUse",
                    "@type": "@id",
                },
                "verifiableCredential": {
                    "@id": "https://www.w3.org/2018/credentials#verifiableCredential",
                    "@type": "@id",
                    "@container": "@graph",
                    "@context": None,
                },
            },
        },
        "EnvelopedVerifiablePresentation": "https://www.w3.org/2018/credentials#EnvelopedVerifiablePresentation",
        "JsonSchemaCredential": "https://www.w3.org/2018/credentials#JsonSchemaCredential",
        "JsonSchema": {
            "@id": "https://www.w3.org/2018/credentials#JsonSchema",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "jsonSchema": {
                    "@id": "https://www.w3.org/2018/credentials#jsonSchema",
                    "@type": "@json",
                },
            },
        },
        "BitstringStatusListCredential": "https://www.w3.org/ns/credentials/status#BitstringStatusListCredential",
        "BitstringStatusList": {
            "@id": "https://www.w3.org/ns/credentials/status#BitstringStatusList",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "encodedList": {
                    "@id": "https://www.w3.org/ns/credentials/status#encodedList",
                    "@type": "https://w3id.org/security#multibase",
                },
                "statusMessage": {
                    "@id": "https://www.w3.org/ns/credentials/status#statusMessage",
                    "@context": {
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "message": "https://www.w3.org/ns/credentials/status#message",
                        "status": "https://www.w3.org/ns/credentials/status#status",
                    },
                },
                "statusPurpose": "https://www.w3.org/ns/credentials/status#statusPurpose",
                "statusReference": {
                    "@id": "https://www.w3.org/ns/credentials/status#statusReference",
                    "@type": "@id",
                },
                "statusSize": {
                    "@id": "https://www.w3.org/ns/credentials/status#statusSize",
                    "@type": "https://www.w3.org/2001/XMLSchema#positiveInteger",
                },
                "ttl": "https://www.w3.org/ns/credentials/status#ttl",
            },
        },
        "BitstringStatusListEntry": {
            "@id": "https://www.w3.org/ns/credentials/status#BitstringStatusListEntry",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "statusListCredential": {
                    "@id": "https://www.w3.org/ns/credentials/status#statusListCredential",
                    "@type": "@id",
                },
                "statusListIndex": "https://www.w3.org/ns/credentials/status#statusListIndex",
                "statusPurpose": "https://www.w3.org/ns/credentials/status#statusPurpose",
            },
        },
        "DataIntegrityProof": {
            "@id": "https://w3id.org/security#DataIntegrityProof",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "challenge": "https://w3id.org/security#challenge",
                "created": {
                    "@id": "http://purl.org/dc/terms/created",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "cryptosuite": {
                    "@id": "https://w3id.org/security#cryptosuite",
                    "@type": "https://w3id.org/security#cryptosuiteString",
                },
                "domain": "https://w3id.org/security#domain",
                "expires": {
                    "@id": "https://w3id.org/security#expiration",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "nonce": "https://w3id.org/security#nonce",
                "previousProof": {
                    "@id": "https://w3id.org/security#previousProof",
                    "@type": "@id",
                },
                "proofPurpose": {
                    "@id": "https://w3id.org/security#proofPurpose",
                    "@type": "@vocab",
                    "@context": {
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "assertionMethod": {
                            "@id": "https://w3id.org/security#assertionMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "authentication": {
                            "@id": "https://w3id.org/security#authenticationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "capabilityDelegation": {
                            "@id": "https://w3id.org/security#capabilityDelegationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "capabilityInvocation": {
                            "@id": "https://w3id.org/security#capabilityInvocationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "keyAgreement": {
                            "@id": "https://w3id.org/security#keyAgreementMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                    },
                },
                "proofValue": {
                    "@id": "https://w3id.org/security#proofValue",
                    "@type": "https://w3id.org/security#multibase",
                },
                "verificationMethod": {
                    "@id": "https://w3id.org/security#verificationMethod",
                    "@type": "@id",
                },
            },
        },
        "...": {"@id": "https://www.iana.org/assignments/jwt#..."},
        "_sd": {"@id": "https://www.iana.org/assignments/jwt#_sd", "@type": "@json"},
        "_sd_alg": {"@id": "https://www.iana.org/assignments/jwt#_sd_alg"},
        "aud": {"@id": "https://www.iana.org/assignments/jwt#aud", "@type": "@id"},
        "cnf": {
            "@id": "https://www.iana.org/assignments/jwt#cnf",
            "@context": {
                "@protected": True,
                "kid": {
                    "@id": "https://www.iana.org/assignments/jwt#kid",
                    "@type": "@id",
                },
                "jwk": {
                    "@id": "https://www.iana.org/assignments/jwt#jwk",
                    "@type": "@json",
                },
            },
        },
        "exp": {
            "@id": "https://www.iana.org/assignments/jwt#exp",
            "@type": "https://www.w3.org/2001/XMLSchema#nonNegativeInteger",
        },
        "iat": {
            "@id": "https://www.iana.org/assignments/jwt#iat",
            "@type": "https://www.w3.org/2001/XMLSchema#nonNegativeInteger",
        },
        "iss": {"@id": "https://www.iana.org/assignments/jose#iss", "@type": "@id"},
        "jku": {"@id": "https://www.iana.org/assignments/jose#jku", "@type": "@id"},
        "kid": {"@id": "https://www.iana.org/assignments/jose#kid", "@type": "@id"},
        "nbf": {
            "@id": "https://www.iana.org/assignments/jwt#nbf",
            "@type": "https://www.w3.org/2001/XMLSchema#nonNegativeInteger",
        },
        "sub": {"@id": "https://www.iana.org/assignments/jose#sub", "@type": "@id"},
        "x5u": {"@id": "https://www.iana.org/assignments/jose#x5u", "@type": "@id"},
    }
}



"""Base message."""

from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import TYPE_CHECKING, Optional, Type

if TYPE_CHECKING:
    from .base_handler import BaseHandler


class DIDCommVersion(Enum):
    """Serialized message formats."""

    v1 = auto()
    v2 = auto()


class BaseMessage(ABC):
    """Abstract base class for messages.

    This formally defines a "minimum viable message" and provides an
    unopinionated class for plugins to extend in whatever way makes sense in
    the context of the plugin.
    """

    @property
    @abstractmethod
    def _type(self) -> str:
        """Return message type."""

    @property
    @abstractmethod
    def _id(self) -> str:
        """Return message id."""

    @property
    @abstractmethod
    def _thread_id(self) -> Optional[str]:
        """Return message thread id."""

    @abstractmethod
    def serialize(self, msg_format: DIDCommVersion = DIDCommVersion.v1) -> dict:
        """Return serialized message in format specified."""

    @classmethod
    @abstractmethod
    def deserialize(cls, value: dict, msg_format: DIDCommVersion = DIDCommVersion.v1):
        """Return message object deserialized from value in format specified."""

    @property
    @abstractmethod
    def Handler(self) -> Type["BaseHandler"]:
        """Return reference to handler class."""



"""Cred request artifacts to attach to RFC 453 messages."""

from typing import Mapping, Optional, Union

from marshmallow import EXCLUDE, fields

from .......messaging.models.base import BaseModel, BaseModelSchema
from .......messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_DID_EXAMPLE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
)


class AnonCredsLinkSecretRequest(BaseModel):
    """Binding proof model."""

    class Meta:
        """VCDI credential request schema metadata."""

        schema_class = "BindingProofSchema"

    def __init__(
        self,
        entropy: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        blinded_ms: Optional[Mapping] = None,
        blinded_ms_correctness_proof: Optional[Mapping] = None,
        nonce: Optional[str] = None,
        **kwargs,
    ):
        """Initialize indy credential request."""
        super().__init__(**kwargs)
        self.entropy = entropy
        self.cred_def_id = cred_def_id
        self.blinded_ms = blinded_ms
        self.blinded_ms_correctness_proof = blinded_ms_correctness_proof
        self.nonce = nonce


class AnonCredsLinkSecretSchema(BaseModelSchema):
    """VCDI credential request schema."""

    class Meta:
        """VCDI credential request schema metadata."""

        model_class = AnonCredsLinkSecretRequest
        unknown = EXCLUDE

    entropy = fields.Str(
        required=True,
        validate=str,
        metadata={"description": "Prover DID", "example": INDY_DID_EXAMPLE},
    )
    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    blinded_ms = fields.Dict(
        required=True, metadata={"description": "Blinded master secret"}
    )
    blinded_ms_correctness_proof = fields.Dict(
        required=True,
        metadata={"description": "Blinded master secret correctness proof"},
    )
    nonce = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Nonce in credential request",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )


class DidcommSignedAttachmentRequest(BaseModel):
    """Didcomm Signed Attachment Model."""

    class Meta:
        """Didcomm signed attachment metadata."""

        schema_class = "DidcommSignedAttachmentSchema"

    def __init__(self, attachment_id: Optional[str] = None, **kwargs):
        """Initialize DidcommSignedAttachment."""
        super().__init__(**kwargs)
        self.attachment_id = attachment_id


class DidcommSignedAttachmentSchema(BaseModelSchema):
    """Didcomm Signed Attachment Schema."""

    class Meta:
        """Didcomm Signed Attachment schema metadata."""

        model_class = DidcommSignedAttachmentRequest
        unknown = EXCLUDE

    attachment_id = fields.Str(required=True, metadata={"description": "", "example": ""})


class BindingProof(BaseModel):
    """Binding Proof Model."""

    class Meta:
        """Binding proof metadata."""

        schema_class = "BindingProofSchema"

    def __init__(
        self,
        anoncreds_link_secret: Optional[str] = None,
        didcomm_signed_attachment: Optional[str] = None,
        **kwargs,
    ):
        """Initialize binding proof."""
        super().__init__(**kwargs)
        self.anoncreds_link_secret = anoncreds_link_secret
        self.didcomm_signed_attachment = didcomm_signed_attachment


class BindingProofSchema(BaseModelSchema):
    """Binding Proof Schema."""

    class Meta:
        """Binding proof schema metadata."""

        model_class = BindingProof
        unknown = EXCLUDE

    anoncreds_link_secret = fields.Nested(
        AnonCredsLinkSecretSchema(),
        required=True,
        metadata={"description": "", "example": ""},
    )

    didcomm_signed_attachment = fields.Nested(
        DidcommSignedAttachmentSchema(),
        required=True,
        metadata={"description": "", "example": ""},
    )


class VCDICredRequest(BaseModel):
    """VCDI credential request model."""

    class Meta:
        """VCDI credential request metadata."""

        schema_class = "VCDICredRequestSchema"

    def __init__(
        self,
        data_model_version: Optional[str] = None,
        binding_proof: Union[dict, BindingProof] = None,
        **kwargs,
    ):
        """Initialize values for VCDICredRequest."""
        super().__init__(**kwargs)
        self.data_model_version = data_model_version
        self.binding_proof = binding_proof


class VCDICredRequestSchema(BaseModelSchema):
    """VCDI credential request schema."""

    class Meta:
        """VCDI credential request schema metadata."""

        model_class = VCDICredRequest
        unknown = EXCLUDE

    data_model_version = fields.Str(
        required=True, metadata={"description": "", "example": ""}
    )

    binding_proof = fields.Nested(
        BindingProofSchema(),
        required=True,
        metadata={"description": "", "example": ""},
    )



TEST_VC_DOCUMENT_NESTED_SIGNED_BBS = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://example.gov/credentials/3732",
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "issuer": "did:example:489398593",
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "credentialSubject": {
        "id": "did:example:489398593",
        "degree": {
            "type": "BachelorDegree",
            "name": "Bachelor of Science and Arts",
            "degreeType": "Underwater Basket Weaving",
        },
        "college": "Contoso University",
    },
    "proof": {
        "type": "BbsBlsSignature2020",
        "created": "2021-04-26T17:55:05Z",
        "proofPurpose": "assertionMethod",
        "proofValue": "t63mgQTFa5eMU0iIpCKtqyuJxC0S7nRE7fJ1g+/YhGnQqTZjAVzdB6mz+B/rEbJFRjoLnV8aJoyX0SXKEZQCZDinxaWksRnSqsDIGz/MVn4ovSqC5zGcApcCidsEO4S77RANaXba6hRHWrjhJGoMPg==",
        "verificationMethod": "did:example:489398593#test",
    },
}



"""Proof verification and validation result classes."""

from typing import Any, List, Optional

from marshmallow import fields

from ...messaging.models.base import BaseModel, BaseModelSchema


class PurposeResult(BaseModel):
    """Proof purpose result class."""

    class Meta:
        """PurposeResult metadata."""

        schema_class = "PurposeResultSchema"

    def __init__(
        self,
        *,
        valid: bool,
        error: Optional[str] = None,
        controller: Optional[Any] = None,
    ) -> None:
        """Create new PurposeResult instance."""
        self.valid = valid
        self.error = error
        self.controller = controller

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        items = ("{}={}".format(k, repr(v)) for k, v in self.__dict__.items())
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))

    def __eq__(self, other: object) -> bool:
        """Comparison between proof purpose results."""
        if isinstance(other, PurposeResult):
            return (
                self.valid == other.valid
                and self.error == other.error
                and self.controller == other.controller
            )
        return False


class PurposeResultSchema(BaseModelSchema):
    """Proof purpose result schema."""

    class Meta:
        """PurposeResultSchema metadata."""

        model_class = PurposeResult

    valid = fields.Boolean()
    error = fields.Str()
    controller = fields.Dict()


class ProofResult(BaseModel):
    """Proof result class."""

    class Meta:
        """ProofResult metadata."""

        schema_class = "ProofResultSchema"

    def __init__(
        self,
        *,
        verified: bool,
        proof: Optional[dict] = None,
        error: Optional[str] = None,
        purpose_result: Optional[PurposeResult] = None,
    ) -> None:
        """Create new ProofResult instance."""
        self.verified = verified
        self.proof = proof
        self.error = error
        self.purpose_result = purpose_result

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        items = ("{}={}".format(k, repr(v)) for k, v in self.__dict__.items())
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))

    def __eq__(self, other: object) -> bool:
        """Comparison between proof results."""
        if isinstance(other, ProofResult):
            return (
                self.verified == other.verified
                and self.proof == other.proof
                and self.error == other.error
                and self.purpose_result == other.purpose_result
            )
        return False


class ProofResultSchema(BaseModelSchema):
    """Proof result schema."""

    class Meta:
        """ProofResultSchema metadata."""

        model_class = ProofResult

    verified = fields.Boolean()
    proof = fields.Dict()
    error = fields.Str()
    purpose_result = fields.Nested(PurposeResultSchema)


class DocumentVerificationResult(BaseModel):
    """Domain verification result class."""

    class Meta:
        """DocumentVerificationResult metadata."""

        schema_class = "DocumentVerificationResultSchema"

    def __init__(
        self,
        *,
        verified: bool,
        document: Optional[dict] = None,
        results: Optional[List[ProofResult]] = None,
        errors: Optional[List[str]] = None,
    ) -> None:
        """Create new DocumentVerificationResult instance."""
        self.verified = verified
        self.document = document
        self.results = results
        self.errors = errors

    def __repr__(self) -> str:
        """Return a human readable representation of this class.

        Returns:
            A human readable string for this class

        """
        items = ("{}={}".format(k, repr(v)) for k, v in self.__dict__.items())
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))

    def __eq__(self, other: object) -> bool:
        """Comparison between document verification results."""
        if isinstance(other, DocumentVerificationResult):
            return (
                self.verified == other.verified
                and self.document == other.document
                # check results list
                and (
                    # both not present
                    (not self.results and not other.results)
                    # both list and matching
                    or (
                        isinstance(self.results, list)
                        and isinstance(other.results, list)
                        and all(
                            self_result == other_result
                            for (self_result, other_result) in zip(
                                self.results, other.results
                            )
                        )
                    )
                )
                # check error list
                and (
                    # both not present
                    (not self.errors and not other.errors)
                    # both list and matching
                    or (
                        isinstance(self.errors, list)
                        and isinstance(other.errors, list)
                        and all(
                            self_error == other_error
                            for (self_error, other_error) in zip(
                                self.errors, other.errors
                            )
                        )
                    )
                )
            )
        return False


class DocumentVerificationResultSchema(BaseModelSchema):
    """Document verification result schema."""

    class Meta:
        """DocumentVerificationResultSchema metadata."""

        model_class = DocumentVerificationResult

    verified = fields.Boolean(required=True)
    document = fields.Dict(required=False)
    results = fields.Nested(ProofResultSchema, many=True)
    errors = fields.List(fields.Str(), required=False)



"""Protocol controller for coordinate mediation."""

from typing import Sequence


class Controller:
    """Coordinate mediation protocol controller."""

    def __init__(self, protocol: str):
        """Initialize the controller."""

    def determine_goal_codes(self) -> Sequence[str]:
        """Return defined goal_codes."""
        return []



from unittest import TestCase

from ..cred_precis import IndyCredInfo

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "preferences"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:default"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:0"

CRED_INFO = {
    "referent": "abc123",
    "attrs": {
        "userid": "alice",
        "dob": "2000-01-01",
        "favourite_colour": "purple",
    },
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "rev_reg_id": REV_REG_ID,
    "cred_rev_id": "1",
}


class TestCredInfo(TestCase):
    """Indy cred info tests"""

    def test_serde(self):
        """Test de/serialization."""
        cred_info = IndyCredInfo.deserialize(CRED_INFO)
        assert type(cred_info) is IndyCredInfo

        ser = cred_info.serialize()
        assert ser == CRED_INFO

        deser = IndyCredInfo.deserialize(ser)
        reser = deser.serialize()
        assert ser == reser

        obj = IndyCredInfo(**CRED_INFO)
        ser = obj.serialize()
        assert ser == reser



from unittest import IsolatedAsyncioTestCase

from ......storage.error import StorageDuplicateError, StorageNotFoundError
from ......tests import mock
from ......utils.testing import create_test_profile
from .....didcomm_prefix import DIDCommPrefix
from ...messages.disclose import Disclose
from ...messages.query import Query
from ..discovery_record import V10DiscoveryExchangeRecord


class TestV10DiscoveryExchangeRecord(IsolatedAsyncioTestCase):
    """Test de/serialization."""

    async def test_record(self):
        same = [
            V10DiscoveryExchangeRecord(
                query_msg=Query(query="*"), connection_id="test123"
            )
        ] * 2
        diff = [
            V10DiscoveryExchangeRecord(
                query_msg=Query(query="test1.*"), connection_id="test123"
            ),
            V10DiscoveryExchangeRecord(
                query_msg=Query(query="test2.*"), connection_id="test123"
            ),
            V10DiscoveryExchangeRecord(
                query_msg=Query(query="test3.*"), connection_id="test123"
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

    async def test_serde(self):
        """Test de/serialization."""
        test_protocols = [
            {
                "pid": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
                "roles": [],
            }
        ]
        query_msg = Query(query="*", comment="test")
        disclose_msg = Disclose(protocols=test_protocols)
        ex_rec = V10DiscoveryExchangeRecord(
            disclose=disclose_msg,
        )
        ex_rec.query_msg = query_msg
        assert isinstance(ex_rec.query_msg, Query)
        ser = ex_rec.serialize()
        deser = V10DiscoveryExchangeRecord.deserialize(ser)
        assert isinstance(deser.query_msg, Query)

        assert isinstance(ex_rec.disclose, Disclose)
        ser = ex_rec.serialize()
        deser = V10DiscoveryExchangeRecord.deserialize(ser)
        assert isinstance(deser.disclose, Disclose)

    async def test_retrieve_by_conn_id(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            record = V10DiscoveryExchangeRecord(
                query_msg=Query(query="*"), connection_id="test123"
            )
            await record.save(session)
            retrieved = await V10DiscoveryExchangeRecord.retrieve_by_connection_id(
                session=session, connection_id="test123"
            )
            assert retrieved
            assert retrieved.connection_id == "test123"

    async def test_exists_for_connection_id(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            record = V10DiscoveryExchangeRecord(
                query_msg=Query(query="*"), connection_id="test123"
            )
            await record.save(session)
            check = await V10DiscoveryExchangeRecord.exists_for_connection_id(
                session=session, connection_id="test123"
            )
            assert check

    async def test_exists_for_connection_id_not_found(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            with mock.patch.object(
                V10DiscoveryExchangeRecord,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(),
            ) as mock_retrieve_by_tag_filter:
                mock_retrieve_by_tag_filter.side_effect = StorageNotFoundError
                check = await V10DiscoveryExchangeRecord.exists_for_connection_id(
                    session=session, connection_id="test123"
                )
                assert not check

    async def test_exists_for_connection_id_duplicate(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            with mock.patch.object(
                V10DiscoveryExchangeRecord,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(),
            ) as mock_retrieve_by_tag_filter:
                mock_retrieve_by_tag_filter.side_effect = StorageDuplicateError
                check = await V10DiscoveryExchangeRecord.exists_for_connection_id(
                    session=session, connection_id="test123"
                )
                assert check



"""Indy issuer implementation."""

import asyncio
import logging
from typing import Optional, Sequence, Tuple

from aries_askar import AskarError
from indy_credx import (
    Credential,
    CredentialDefinition,
    CredentialOffer,
    CredentialRevocationConfig,
    CredxError,
    RevocationRegistry,
    RevocationRegistryDefinition,
    RevocationRegistryDefinitionPrivate,
    RevocationRegistryDelta,
    Schema,
)

from ...askar.profile import AskarProfile
from ...utils.general import strip_did_prefix
from ..issuer import (
    DEFAULT_CRED_DEF_TAG,
    DEFAULT_SIGNATURE_TYPE,
    IndyIssuer,
    IndyIssuerError,
    IndyIssuerRevocationRegistryFullError,
)

LOGGER = logging.getLogger(__name__)

CATEGORY_CRED_DEF = "credential_def"
CATEGORY_CRED_DEF_PRIVATE = "credential_def_private"
CATEGORY_CRED_DEF_KEY_PROOF = "credential_def_key_proof"
CATEGORY_SCHEMA = "schema"
CATEGORY_REV_REG = "revocation_reg"
CATEGORY_REV_REG_INFO = "revocation_reg_info"
CATEGORY_REV_REG_DEF = "revocation_reg_def"
CATEGORY_REV_REG_DEF_PRIVATE = "revocation_reg_def_private"
CATEGORY_REV_REG_ISSUER = "revocation_reg_def_issuer"


class IndyCredxIssuer(IndyIssuer):
    """Indy-Credx issuer class."""

    def __init__(self, profile: AskarProfile):
        """Initialize an IndyCredxIssuer instance.

        Args:
            profile: The active profile instance

        """
        self._profile = profile

    @property
    def profile(self) -> AskarProfile:
        """Accessor for the profile instance."""
        return self._profile

    async def create_schema(
        self,
        origin_did: str,
        schema_name: str,
        schema_version: str,
        attribute_names: Sequence[str],
    ) -> Tuple[str, str]:
        """Create a new credential schema and store it in the wallet.

        Args:
            origin_did: the DID issuing the credential definition
            schema_name: the schema name
            schema_version: the schema version
            attribute_names: a sequence of schema attribute names

        Returns:
            A tuple of the schema ID and JSON

        """
        try:
            schema = Schema.create(
                strip_did_prefix(origin_did),
                schema_name,
                schema_version,
                attribute_names,
            )
            schema_id = schema.id
            schema_json = schema.to_json()
            async with self._profile.session() as session:
                await session.handle.insert(CATEGORY_SCHEMA, schema_id, schema_json)
        except CredxError as err:
            raise IndyIssuerError("Error creating schema") from err
        except AskarError as err:
            raise IndyIssuerError("Error storing schema") from err
        return (schema_id, schema_json)

    async def credential_definition_in_wallet(
        self, credential_definition_id: str
    ) -> bool:
        """Check whether a given credential definition ID is present in the wallet.

        Args:
            credential_definition_id: The credential definition ID to check
        """
        try:
            async with self._profile.session() as session:
                return (
                    await session.handle.fetch(
                        CATEGORY_CRED_DEF_PRIVATE, credential_definition_id
                    )
                ) is not None
        except AskarError as err:
            raise IndyIssuerError("Error checking for credential definition") from err

    async def create_and_store_credential_definition(
        self,
        origin_did: str,
        schema: dict,
        signature_type: Optional[str] = None,
        tag: Optional[str] = None,
        support_revocation: bool = False,
    ) -> Tuple[str, str]:
        """Create a new credential definition and store it in the wallet.

        Args:
            origin_did (str): The DID issuing the credential definition.
            schema (dict): The schema to create a credential definition for.
            signature_type (str, optional): The credential definition signature type
                (default 'CL').
            tag (str, optional): The credential definition tag.
            support_revocation (bool, optional): Whether to enable revocation for this
                credential definition.

        Returns:
            Tuple[str, str]: A tuple of the credential definition ID and JSON.

        Raises:
            IndyIssuerError: If there is an error creating or storing the credential
                definition.

        """
        try:
            (
                cred_def,
                cred_def_private,
                key_proof,
            ) = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: CredentialDefinition.create(
                    strip_did_prefix(origin_did),
                    schema,
                    signature_type or DEFAULT_SIGNATURE_TYPE,
                    tag or DEFAULT_CRED_DEF_TAG,
                    support_revocation=support_revocation,
                ),
            )
            cred_def_id = cred_def.id
            cred_def_json = cred_def.to_json()
        except CredxError as err:
            raise IndyIssuerError("Error creating credential definition") from err
        try:
            async with self._profile.transaction() as txn:
                await txn.handle.insert(
                    CATEGORY_CRED_DEF,
                    cred_def_id,
                    cred_def_json,
                    # Note: Indy-SDK uses a separate SchemaId record for this
                    tags={"schema_id": schema["id"]},
                )
                await txn.handle.insert(
                    CATEGORY_CRED_DEF_PRIVATE,
                    cred_def_id,
                    cred_def_private.to_json_buffer(),
                )
                await txn.handle.insert(
                    CATEGORY_CRED_DEF_KEY_PROOF, cred_def_id, key_proof.to_json_buffer()
                )
                await txn.commit()
        except AskarError as err:
            raise IndyIssuerError("Error storing credential definition") from err
        return (cred_def_id, cred_def_json)

    async def create_credential_offer(self, credential_definition_id: str) -> str:
        """Create a credential offer for the given credential definition id.

        Args:
            credential_definition_id: The credential definition to create an offer for

        Returns:
            The new credential offer

        """
        try:
            async with self._profile.session() as session:
                cred_def = await session.handle.fetch(
                    CATEGORY_CRED_DEF, credential_definition_id
                )
                key_proof = await session.handle.fetch(
                    CATEGORY_CRED_DEF_KEY_PROOF, credential_definition_id
                )
        except AskarError as err:
            raise IndyIssuerError("Error retrieving credential definition") from err
        if not cred_def or not key_proof:
            raise IndyIssuerError("Credential definition not found for credential offer")
        try:
            # The tag holds the full name of the schema,
            # as opposed to just the sequence number
            schema_id = cred_def.tags.get("schema_id")
            cred_def = CredentialDefinition.load(cred_def.raw_value)

            credential_offer = CredentialOffer.create(
                schema_id or cred_def.schema_id,
                cred_def,
                key_proof.raw_value,
            )
        except CredxError as err:
            raise IndyIssuerError("Error creating credential offer") from err

        return credential_offer.to_json()

    async def create_credential(
        self,
        schema: dict,
        credential_offer: dict,
        credential_request: dict,
        credential_values: dict,
        revoc_reg_id: Optional[str] = None,
        tails_file_path: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Create a credential.

        Args:
            schema: Schema to create credential for
            credential_offer: Credential Offer to create credential for
            credential_request: Credential request to create credential for
            credential_values: Values to go in credential
            revoc_reg_id: ID of the revocation registry
            tails_file_path: The location of the tails file

        Returns:
            A tuple of created credential and revocation id

        """
        credential_definition_id = credential_offer["cred_def_id"]
        try:
            async with self._profile.session() as session:
                cred_def = await session.handle.fetch(
                    CATEGORY_CRED_DEF, credential_definition_id
                )
                cred_def_private = await session.handle.fetch(
                    CATEGORY_CRED_DEF_PRIVATE, credential_definition_id
                )
        except AskarError as err:
            raise IndyIssuerError("Error retrieving credential definition") from err
        if not cred_def or not cred_def_private:
            raise IndyIssuerError(
                "Credential definition not found for credential issuance"
            )

        raw_values = {}
        schema_attributes = schema["attrNames"]
        for attribute in schema_attributes:
            # Ensure every attribute present in schema to be set.
            # Extraneous attribute names are ignored.
            try:
                credential_value = credential_values[attribute]
            except KeyError:
                raise IndyIssuerError(
                    "Provided credential values are missing a value "
                    f"for the schema attribute '{attribute}'"
                )

            raw_values[attribute] = str(credential_value)

        if revoc_reg_id:
            try:
                async with self._profile.transaction() as txn:
                    rev_reg = await txn.handle.fetch(CATEGORY_REV_REG, revoc_reg_id)
                    rev_reg_info = await txn.handle.fetch(
                        CATEGORY_REV_REG_INFO, revoc_reg_id, for_update=True
                    )
                    rev_reg_def = await txn.handle.fetch(
                        CATEGORY_REV_REG_DEF, revoc_reg_id
                    )
                    rev_key = await txn.handle.fetch(
                        CATEGORY_REV_REG_DEF_PRIVATE, revoc_reg_id
                    )
                    if not rev_reg:
                        raise IndyIssuerError("Revocation registry not found")
                    if not rev_reg_info:
                        raise IndyIssuerError("Revocation registry metadata not found")
                    if not rev_reg_def:
                        raise IndyIssuerError("Revocation registry definition not found")
                    if not rev_key:
                        raise IndyIssuerError(
                            "Revocation registry definition private data not found"
                        )
                    # NOTE: we increment the index ahead of time to keep the
                    # transaction short. The revocation registry itself will NOT
                    # be updated because we always use ISSUANCE_BY_DEFAULT.
                    # If something goes wrong later, the index will be skipped.
                    # FIXME - double check issuance type in case of upgraded wallet?
                    rev_info = rev_reg_info.value_json
                    rev_reg_index = rev_info["curr_id"] + 1
                    try:
                        rev_reg_def = RevocationRegistryDefinition.load(
                            rev_reg_def.raw_value
                        )
                    except CredxError as err:
                        raise IndyIssuerError(
                            "Error loading revocation registry definition"
                        ) from err
                    if rev_reg_index > rev_reg_def.max_cred_num:
                        raise IndyIssuerRevocationRegistryFullError(
                            "Revocation registry is full"
                        )
                    rev_info["curr_id"] = rev_reg_index
                    await txn.handle.replace(
                        CATEGORY_REV_REG_INFO, revoc_reg_id, value_json=rev_info
                    )
                    await txn.commit()
            except AskarError as err:
                raise IndyIssuerError("Error updating revocation registry index") from err

            revoc = CredentialRevocationConfig(
                rev_reg_def,
                rev_key.raw_value,
                rev_reg.raw_value,
                rev_reg_index,
                rev_info.get("used_ids") or [],
            )
            credential_revocation_id = str(rev_reg_index)
        else:
            revoc = None
            credential_revocation_id = None

        # This is for compatibility with an anoncreds holder
        if not credential_request.get("prover_did"):
            credential_request["prover_did"] = credential_request["entropy"]
            del credential_request["entropy"]

        try:
            (
                credential,
                _upd_rev_reg,
                _delta,
            ) = await asyncio.get_event_loop().run_in_executor(
                None,
                Credential.create,
                cred_def.raw_value,
                cred_def_private.raw_value,
                credential_offer,
                credential_request,
                raw_values,
                None,
                revoc,
            )
        except CredxError as err:
            raise IndyIssuerError("Error creating credential") from err

        return credential.to_json(), credential_revocation_id

    async def revoke_credentials(
        self,
        cred_def_id: str,
        revoc_reg_id: str,
        tails_file_path: str,
        cred_revoc_ids: Sequence[str],
    ) -> Tuple[str, Sequence[str]]:
        """Revoke a set of credentials in a revocation registry.

        Args:
            cred_def_id: ID of the credential definition
            revoc_reg_id: ID of the revocation registry
            tails_file_path: path to the local tails file
            cred_revoc_ids: sequences of credential indexes in the revocation registry

        Returns:
            Tuple with the combined revocation delta, list of cred rev ids not revoked

        """

        delta = None
        failed_crids = set()
        max_attempt = 5
        attempt = 0

        while True:
            attempt += 1
            if attempt >= max_attempt:
                raise IndyIssuerError("Repeated conflict attempting to update registry")
            try:
                async with self._profile.session() as session:
                    cred_def = await session.handle.fetch(CATEGORY_CRED_DEF, cred_def_id)
                    rev_reg_def = await session.handle.fetch(
                        CATEGORY_REV_REG_DEF, revoc_reg_id
                    )
                    rev_reg_def_private = await session.handle.fetch(
                        CATEGORY_REV_REG_DEF_PRIVATE, revoc_reg_id
                    )
                    rev_reg = await session.handle.fetch(CATEGORY_REV_REG, revoc_reg_id)
                    rev_reg_info = await session.handle.fetch(
                        CATEGORY_REV_REG_INFO, revoc_reg_id
                    )
                if not cred_def:
                    raise IndyIssuerError("Credential definition not found")
                if not rev_reg_def:
                    raise IndyIssuerError("Revocation registry definition not found")
                if not rev_reg_def_private:
                    raise IndyIssuerError(
                        "Revocation registry definition private key not found"
                    )
                if not rev_reg:
                    raise IndyIssuerError("Revocation registry not found")
                if not rev_reg_info:
                    raise IndyIssuerError("Revocation registry metadata not found")
            except AskarError as err:
                raise IndyIssuerError("Error retrieving revocation registry") from err

            try:
                cred_def = CredentialDefinition.load(cred_def.raw_value)
            except CredxError as err:
                raise IndyIssuerError("Error loading credential definition") from err

            try:
                rev_reg_def = RevocationRegistryDefinition.load(rev_reg_def.raw_value)
            except CredxError as err:
                raise IndyIssuerError(
                    "Error loading revocation registry definition"
                ) from err

            try:
                rev_reg_def_private = RevocationRegistryDefinitionPrivate.load(
                    rev_reg_def_private.raw_value
                )
            except CredxError as err:
                raise IndyIssuerError(
                    "Error loading revocation registry private key"
                ) from err

            try:
                rev_reg = RevocationRegistry.load(rev_reg.raw_value)
            except CredxError as err:
                raise IndyIssuerError("Error loading revocation registry") from err

            rev_crids = set()
            failed_crids = set()
            max_cred_num = rev_reg_def.max_cred_num
            rev_info = rev_reg_info.value_json
            used_ids = set(rev_info.get("used_ids") or [])

            for rev_id in cred_revoc_ids:
                rev_id = int(rev_id)
                if rev_id < 1 or rev_id > max_cred_num:
                    LOGGER.error(
                        "Skipping requested credential revocation"
                        "on rev reg id %s, cred rev id=%s not in range",
                        revoc_reg_id,
                        rev_id,
                    )
                    failed_crids.add(rev_id)
                elif rev_id > rev_info["curr_id"]:
                    LOGGER.warning(
                        "Skipping requested credential revocation"
                        "on rev reg id %s, cred rev id=%s not yet issued",
                        revoc_reg_id,
                        rev_id,
                    )
                    failed_crids.add(rev_id)
                elif rev_id in used_ids:
                    LOGGER.warning(
                        "Skipping requested credential revocation"
                        "on rev reg id %s, cred rev id=%s already revoked",
                        revoc_reg_id,
                        rev_id,
                    )
                    failed_crids.add(rev_id)
                else:
                    rev_crids.add(rev_id)

            if not rev_crids:
                break

            try:
                delta = await asyncio.get_event_loop().run_in_executor(
                    None,
                    lambda: rev_reg.update(
                        cred_def,
                        rev_reg_def,
                        rev_reg_def_private,
                        issued=None,
                        revoked=list(rev_crids),  # revoked
                    ),
                )
            except CredxError as err:
                raise IndyIssuerError("Error updating revocation registry") from err

            try:
                async with self._profile.transaction() as txn:
                    rev_reg_upd = await txn.handle.fetch(
                        CATEGORY_REV_REG, revoc_reg_id, for_update=True
                    )
                    rev_info_upd = await txn.handle.fetch(
                        CATEGORY_REV_REG_INFO, revoc_reg_id, for_update=True
                    )
                    if not rev_reg_upd or not rev_reg_info:
                        LOGGER.warning(
                            "Revocation registry missing, skipping update: {}",
                            revoc_reg_id,
                        )
                        delta = None
                        break
                    rev_info_upd = rev_info_upd.value_json
                    if rev_info_upd != rev_info:
                        # handle concurrent update to the registry by retrying
                        continue
                    await txn.handle.replace(
                        CATEGORY_REV_REG, revoc_reg_id, rev_reg.to_json_buffer()
                    )
                    used_ids.update(rev_crids)
                    rev_info_upd["used_ids"] = sorted(used_ids)
                    await txn.handle.replace(
                        CATEGORY_REV_REG_INFO, revoc_reg_id, value_json=rev_info_upd
                    )
                    await txn.commit()
            except AskarError as err:
                raise IndyIssuerError("Error saving revocation registry") from err
            break

        return (
            delta and delta.to_json(),
            [str(rev_id) for rev_id in sorted(failed_crids)],
        )

    async def merge_revocation_registry_deltas(
        self, fro_delta: str, to_delta: str
    ) -> str:
        """Merge revocation registry deltas.

        Args:
            fro_delta: original delta in JSON format
            to_delta: incoming delta in JSON format

        Returns:
            Merged delta in JSON format

        """

        def update(d1, d2):
            try:
                delta = RevocationRegistryDelta.load(d1)
                delta.update_with(d2)
                return delta.to_json()
            except CredxError as err:
                raise IndyIssuerError("Error merging revocation registry deltas") from err

        return await asyncio.get_event_loop().run_in_executor(
            None, update, fro_delta, to_delta
        )

    async def create_and_store_revocation_registry(
        self,
        origin_did: str,
        cred_def_id: str,
        revoc_def_type: str,
        tag: str,
        max_cred_num: int,
        tails_base_path: str,
    ) -> Tuple[str, str, str]:
        """Create a new revocation registry and store it in the wallet.

        Args:
            origin_did: the DID issuing the revocation registry
            cred_def_id: the identifier of the related credential definition
            revoc_def_type: the revocation registry type (default CL_ACCUM)
            tag: the unique revocation registry tag
            max_cred_num: the number of credentials supported in the registry
            tails_base_path: where to store the tails file
            issuance_type: optionally override the issuance type

        Returns:
            A tuple of the revocation registry ID, JSON, and entry JSON

        """
        try:
            async with self._profile.session() as session:
                cred_def = await session.handle.fetch(CATEGORY_CRED_DEF, cred_def_id)
        except AskarError as err:
            raise IndyIssuerError("Error retrieving credential definition") from err
        if not cred_def:
            raise IndyIssuerError(
                "Credential definition not found for revocation registry"
            )

        try:
            (
                rev_reg_def,
                rev_reg_def_private,
                rev_reg,
                _rev_reg_delta,
            ) = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: RevocationRegistryDefinition.create(
                    strip_did_prefix(origin_did),
                    cred_def.raw_value,
                    tag,
                    revoc_def_type,
                    max_cred_num,
                    tails_dir_path=tails_base_path,
                ),
            )
        except CredxError as err:
            raise IndyIssuerError("Error creating revocation registry") from err

        rev_reg_def_id = rev_reg_def.id
        rev_reg_def_json = rev_reg_def.to_json()
        rev_reg_json = rev_reg.to_json()

        try:
            async with self._profile.transaction() as txn:
                await txn.handle.insert(CATEGORY_REV_REG, rev_reg_def_id, rev_reg_json)
                await txn.handle.insert(
                    CATEGORY_REV_REG_INFO,
                    rev_reg_def_id,
                    value_json={"curr_id": 0, "used_ids": []},
                )
                await txn.handle.insert(
                    CATEGORY_REV_REG_DEF, rev_reg_def_id, rev_reg_def_json
                )
                await txn.handle.insert(
                    CATEGORY_REV_REG_DEF_PRIVATE,
                    rev_reg_def_id,
                    rev_reg_def_private.to_json_buffer(),
                )
                await txn.commit()
        except AskarError as err:
            raise IndyIssuerError("Error saving new revocation registry") from err

        return (
            rev_reg_def_id,
            rev_reg_def_json,
            rev_reg_json,
        )



"""Constants for State Proof and LeafHash Inclusion Verification."""

ALL_ATR_KEYS = ["raw", "enc", "hash"]
LAST_SEQ_NO = "lsn"
VALUE = "value"
VAL = "val"
HASH = "hash"
LAST_UPDATE_TIME = "lut"
MARKER_CLAIM_DEF = "3"
MARKER_SCHEMA = "2"
MARKER_ATTR = "1"
MARKER_REVOC_DEF = "4"
MARKER_REVOC_REG_ENTRY_ACCUM = "6"
MARKER_REVOC_REG_ENTRY = "5"
GET_NYM = "105"
GET_ATTR = "104"
GET_SCHEMA = "107"
GET_CLAIM_DEF = "108"
GET_REVOC_REG_DEF = "115"
GET_REVOC_REG_ENTRY = "116"
GET_REVOC_REG_DELTA = "117"
NYM = "1"
ATTRIB = "100"
SCHEMA = "101"
CLAIM_DEF = "102"
REVOC_REG_DEF = "113"
REVOC_REG_ENTRY = "114"
DEST = "dest"
RESULT = "result"
DATA = "data"
SEQ_NO = "seqNo"
TXN_TIME = "txnTime"
TXN_METADATA = "txnMetadata"
TXN_TIME = "txnTime"
TXN = "txn"
NAME = "name"
VERSION = "version"
ATTR_NAMES = "attr_names"
FROM = "from"
METADATA = "metadata"
ORIGIN = "origin"
REF = "ref"
IDENTIFIER = "identifier"
CRED_DEF_ID = "credDefId"
REVOC_DEF_TYPE = "revocDefType"
REVOC_DEF_TYPE_ID = "revocRegDefId"
AUDIT_PATH = "auditPath"
ROOT_HASH = "rootHash"
STATE_PROOF = "state_proof"
STATE_PROOF_FROM = "stateProofFrom"
PROOF_NODES = "proof_nodes"
TAG = "tag"
ACCUM_TO = "accum_to"
ACCUM_FROM = "accum_from"
(NODE_TYPE_BLANK, NODE_TYPE_LEAF, NODE_TYPE_EXTENSION, NODE_TYPE_BRANCH) = tuple(range(4))
BLANK_NODE = b""



"""VC-API Routes."""

from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema
from marshmallow.exceptions import ValidationError
from uuid_utils import uuid4

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..config.base import InjectionError
from ..resolver.base import ResolverError
from ..storage.error import StorageDuplicateError, StorageError, StorageNotFoundError
from ..storage.vc_holder.base import VCHolder
from ..wallet.base import BaseWallet
from ..wallet.error import WalletError
from .vc_ld.manager import VcLdpManager, VcLdpManagerError
from .vc_ld.models import web_schemas
from .vc_ld.models.credential import VerifiableCredential
from .vc_ld.models.options import LDProofVCOptions
from .vc_ld.models.presentation import VerifiablePresentation


@docs(tags=["vc-api"], summary="List credentials")
@response_schema(web_schemas.ListCredentialsResponse(), 200, description="")
@tenant_authentication
async def list_credentials_route(request: web.BaseRequest):
    """Request handler for issuing a credential.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    holder = context.inject(VCHolder)
    try:
        search = holder.search_credentials()
        records = {
            "results": [
                record.serialize()["cred_value"] for record in await search.fetch()
            ]
        }
        return web.json_response(records, status=200)
    except (StorageError, StorageNotFoundError) as err:
        return web.json_response({"message": err.roll_up}, status=400)


@docs(tags=["vc-api"], summary="Fetch credential by ID")
@response_schema(web_schemas.FetchCredentialResponse(), 200, description="")
@tenant_authentication
async def fetch_credential_route(request: web.BaseRequest):
    """Request handler for issuing a credential.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    holder = context.inject(VCHolder)
    try:
        credential_id = request.match_info["credential_id"].strip('"')
        record = await holder.retrieve_credential_by_id(record_id=credential_id)
        return web.json_response(record.serialize()["cred_value"], status=200)
    except (StorageError, StorageNotFoundError) as err:
        return web.json_response({"message": err.roll_up}, status=400)


@docs(tags=["vc-api"], summary="Issue a credential")
@request_schema(web_schemas.IssueCredentialRequest())
@response_schema(web_schemas.IssueCredentialResponse(), 200, description="")
@tenant_authentication
async def issue_credential_route(request: web.BaseRequest):
    """Request handler for issuing a credential.

    Args:
        request: aiohttp request object

    """
    body = await request.json()
    context: AdminRequestContext = request["context"]
    manager = VcLdpManager(context.profile)
    try:
        credential = body["credential"]
        options = {} if "options" not in body else body["options"]

        # We derive the proofType from the issuer DID if not provided in options
        if not options.get("proofType", None):
            issuer = credential["issuer"]
            did = issuer if isinstance(issuer, str) else issuer["id"]
            async with context.session() as session:
                wallet: BaseWallet | None = session.inject_or(BaseWallet)
                info = await wallet.get_local_did(did)
                key_type = info.key_type.key_type

            if key_type == "ed25519":
                options["proofType"] = "Ed25519Signature2020"
            elif key_type == "bls12381g2":
                options["proofType"] = "BbsBlsSignature2020"
            elif key_type == "p256":
                options["proofType"] = "EcdsaSecp256r1Signature2019"

        credential = VerifiableCredential.deserialize(credential)
        options = LDProofVCOptions.deserialize(options)

        vc = await manager.issue(credential, options)
        response = {"verifiableCredential": vc.serialize()}
        return web.json_response(response, status=201)
    except (ValidationError, VcLdpManagerError, WalletError, InjectionError) as err:
        return web.json_response({"message": str(err)}, status=400)


@docs(tags=["vc-api"], summary="Verify a credential")
@request_schema(web_schemas.VerifyCredentialRequest())
@response_schema(web_schemas.VerifyCredentialResponse(), 200, description="")
@tenant_authentication
async def verify_credential_route(request: web.BaseRequest):
    """Request handler for verifying a credential.

    Args:
        request: aiohttp request object

    """
    body = await request.json()
    context: AdminRequestContext = request["context"]
    manager = VcLdpManager(context.profile)
    try:
        vc = VerifiableCredential.deserialize(body["verifiableCredential"])
        result = await manager.verify_credential(vc)
        result = result.serialize()
        return web.json_response(result)
    except (
        ValidationError,
        VcLdpManagerError,
        ResolverError,
        ValueError,
        WalletError,
        InjectionError,
    ) as err:
        return web.json_response({"message": str(err)}, status=400)


@docs(tags=["vc-api"], summary="Store a credential")
@request_schema(web_schemas.StoreCredentialRequest())
@response_schema(web_schemas.StoreCredentialResponse(), 200, description="")
@tenant_authentication
async def store_credential_route(request: web.BaseRequest):
    """Request handler for storing a credential.

    Args:
        request: aiohttp request object

    """
    body = await request.json()
    context: AdminRequestContext = request["context"]
    manager = VcLdpManager(context.profile)

    try:
        vc = body["verifiableCredential"]
        cred_id = vc["id"] if "id" in vc else f"urn:uuid:{str(uuid4())}"
        options = {} if "options" not in body else body["options"]

        vc = VerifiableCredential.deserialize(vc)
        options = LDProofVCOptions.deserialize(options)

        await manager.verify_credential(vc)
        await manager.store_credential(vc, cred_id)

        return web.json_response({"credentialId": cred_id}, status=200)

    except (
        ValidationError,
        VcLdpManagerError,
        WalletError,
        InjectionError,
        StorageDuplicateError,
    ) as err:
        return web.json_response({"message": str(err)}, status=400)


@docs(tags=["vc-api"], summary="Prove a presentation")
@request_schema(web_schemas.ProvePresentationRequest())
@response_schema(web_schemas.ProvePresentationResponse(), 200, description="")
@tenant_authentication
async def prove_presentation_route(request: web.BaseRequest):
    """Request handler for proving a presentation.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    manager = VcLdpManager(context.profile)
    body = await request.json()
    try:
        presentation = body["presentation"]
        options = {} if "options" not in body else body["options"]

        # We derive the proofType from the holder DID if not provided in options
        if not options.get("proofType", None) and presentation.get("holder"):
            holder = presentation["holder"]
            did = holder if isinstance(holder, str) else holder["id"]
            async with context.session() as session:
                wallet: BaseWallet | None = session.inject_or(BaseWallet)
                info = await wallet.get_local_did(did)
                key_type = info.key_type.key_type

            if key_type == "ed25519":
                options["proofType"] = "Ed25519Signature2020"
            elif key_type == "bls12381g2":
                options["proofType"] = "BbsBlsSignature2020"
            elif key_type == "p256":
                options["proofType"] = "EcdsaSecp256r1Signature2019"

        else:
            options["proofType"] = options.get("proofType") or "Ed25519Signature2020"

        presentation = VerifiablePresentation.deserialize(presentation)
        options = LDProofVCOptions.deserialize(options)
        vp = await manager.prove(presentation, options)
        return web.json_response({"verifiablePresentation": vp.serialize()}, status=201)

    except (ValidationError, VcLdpManagerError, WalletError, InjectionError) as err:
        return web.json_response({"message": str(err)}, status=400)


@docs(tags=["vc-api"], summary="Verify a Presentation")
@request_schema(web_schemas.VerifyPresentationRequest())
@response_schema(web_schemas.VerifyPresentationResponse(), 200, description="")
@tenant_authentication
async def verify_presentation_route(request: web.BaseRequest):
    """Request handler for verifying a presentation.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    manager = VcLdpManager(context.profile)
    body = await request.json()
    try:
        vp = VerifiablePresentation.deserialize(body["verifiablePresentation"])
        options = {} if "options" not in body else body["options"]
        options = LDProofVCOptions.deserialize(options)
        verified = await manager.verify_presentation(vp, options)
        return web.json_response(verified.serialize(), status=200)
    except (
        ValidationError,
        WalletError,
        InjectionError,
        VcLdpManagerError,
        ResolverError,
        ValueError,
    ) as err:
        return web.json_response({"message": str(err)}, status=400)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/vc/credentials", list_credentials_route, allow_head=False),
            web.get(
                "/vc/credentials/{credential_id}",
                fetch_credential_route,
                allow_head=False,
            ),
            web.post("/vc/credentials/issue", issue_credential_route),
            web.post("/vc/credentials/store", store_credential_route),
            web.post("/vc/credentials/verify", verify_credential_route),
            web.post("/vc/presentations/prove", prove_presentation_route),
            web.post("/vc/presentations/verify", verify_presentation_route),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""
    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "vc-api",
            "description": "Endpoints for managing w3c credentials and presentations",
            "externalDocs": {
                "description": "Specification",
                "url": "https://w3c-ccg.github.io/vc-api/",
            },
        }
    )



"""Oob message processor and functions."""

import json
import logging
from typing import Any, Callable, Dict, List, Optional

from ..connections.models.conn_record import ConnRecord
from ..connections.models.connection_target import ConnectionTarget
from ..messaging.agent_message import AgentMessage
from ..messaging.decorators.service_decorator import ServiceDecorator
from ..messaging.request_context import RequestContext
from ..protocols.didcomm_prefix import DIDCommPrefix
from ..protocols.issue_credential.v1_0.message_types import CREDENTIAL_OFFER
from ..protocols.issue_credential.v2_0.message_types import CRED_20_OFFER
from ..protocols.out_of_band.v1_0.models.oob_record import OobRecord
from ..protocols.present_proof.v1_0.message_types import PRESENTATION_REQUEST
from ..protocols.present_proof.v2_0.message_types import PRES_20_REQUEST
from ..storage.error import StorageNotFoundError
from ..transport.inbound.message import InboundMessage
from ..transport.outbound.message import OutboundMessage
from ..transport.wire_format import JsonWireFormat
from .error import BaseError
from .profile import Profile

LOGGER = logging.getLogger(__name__)


class OobMessageProcessorError(BaseError):
    """Base error for OobMessageProcessor."""


class OobMessageProcessor:
    """Out of band message processor."""

    def __init__(
        self,
        inbound_message_router: Callable[[Profile, InboundMessage, Optional[bool]], None],
    ) -> None:
        """Initialize an inbound OOB message processor.

        Args:
            inbound_message_router: Method to create a new inbound session

        """
        self._inbound_message_router = inbound_message_router
        self.wire_format = JsonWireFormat()

    async def clean_finished_oob_record(self, profile: Profile, message: AgentMessage):
        """Clean up oob record associated with agent message, if applicable."""
        try:
            async with profile.session() as session:
                oob_record = await OobRecord.retrieve_by_tag_filter(
                    session,
                    {"invi_msg_id": message._thread.pthid},
                    {"role": OobRecord.ROLE_SENDER},
                )

                # emit the "done" event for OOB invitations
                if not oob_record.invitation.requests_attach:
                    oob_record.state = OobRecord.STATE_DONE
                    await oob_record.emit_event(session)

                # If the oob record is not multi use and it doesn't contain any
                # attachments, we can now safely remove the oob record
                if not oob_record.multi_use and not oob_record.invitation.requests_attach:
                    await oob_record.delete_record(session)
        except StorageNotFoundError:
            # It is fine if no oob record is found, Only retrieved for cleanup
            pass
        except Exception:
            LOGGER.warning("Error cleaning up oob record", exc_info=True)

    async def find_oob_target_for_outbound_message(
        self, profile: Profile, outbound_message: OutboundMessage
    ) -> Optional[ConnectionTarget]:
        """Find connection target for the outbound message."""
        try:
            async with profile.session() as session:
                # Try to find the oob record for the outbound message:
                oob_record = await OobRecord.retrieve_by_tag_filter(
                    session, {"attach_thread_id": outbound_message.reply_thread_id}
                )

                LOGGER.debug(
                    "extracting their service from oob record %s",
                    oob_record.their_service,
                )

                their_service = oob_record.their_service
                if not their_service:
                    raise OobMessageProcessorError("Could not determine their service")

                # Attach ~service decorator so other message can respond
                message = json.loads(outbound_message.payload)
                if not message.get("~service") and oob_record.our_service:
                    LOGGER.debug(
                        "Setting our service on the message ~service %s",
                        oob_record.our_service,
                    )
                    message["~service"] = oob_record.our_service.serialize()

                message["~thread"] = {
                    **message.get("~thread", {}),
                    "pthid": oob_record.invi_msg_id,
                }

                outbound_message.payload = json.dumps(message)

                LOGGER.debug("Sending oob message payload %s", outbound_message.payload)

                return ConnectionTarget(
                    endpoint=their_service.endpoint,
                    recipient_keys=their_service.recipient_keys,
                    routing_keys=their_service.routing_keys,
                    sender_key=oob_record.our_recipient_key,
                )
        except StorageNotFoundError:
            return None

    async def find_oob_record_for_inbound_message(
        self, context: RequestContext
    ) -> Optional[OobRecord]:
        """Find oob record for inbound message."""
        message_type = context.message._type
        oob_record = None
        async with context.profile.session() as session:
            # First try to find the oob record based on the associated pthid
            if context.message_receipt.parent_thread_id:
                try:
                    LOGGER.debug(
                        "Retrieving OOB record using pthid "
                        f"{context.message_receipt.parent_thread_id} "
                        f"for message type {message_type}"
                    )
                    oob_record = await OobRecord.retrieve_by_tag_filter(
                        session,
                        {"invi_msg_id": context.message_receipt.parent_thread_id},
                    )
                except StorageNotFoundError:
                    # Fine if record is not found
                    pass
            # Otherwise try to find it using the attach thread id. This is only needed
            # for connectionless exchanges where every handler needs the context of the
            # oob record for verification. We could attach the oob_record to all messages,
            # even if we have a connection, but it would add another query to all inbound
            # messages.
            if (
                not oob_record
                and not context.connection_record
                and context.message_receipt.thread_id
                and context.message_receipt.recipient_verkey
            ):
                try:
                    LOGGER.debug(
                        "Retrieving OOB record using thid "
                        f"{context.message_receipt.thread_id} and recipient verkey"
                        f" {context.message_receipt.recipient_verkey} for "
                        f"message type {message_type}"
                    )
                    oob_record = await OobRecord.retrieve_by_tag_filter(
                        session,
                        {
                            "attach_thread_id": context.message_receipt.thread_id,
                            "our_recipient_key": context.message_receipt.recipient_verkey,
                        },
                    )
                except StorageNotFoundError:
                    # Fine if record is not found
                    pass

        # If not oob record was found we can return early without oob record
        if not oob_record:
            return None

        LOGGER.debug(
            f"Found out of band record for inbound message with type {message_type}"
            f": {oob_record.oob_id}"
        )

        # If the connection does not match with the connection id associated with the
        # oob record we don't want to associate the oob record to the current context
        # This is not the case if the state is await response, in this case we might want
        # to update the connection id on the oob record
        if (
            # Only if we created the invitation
            oob_record.role == OobRecord.ROLE_SENDER
            # If connection is present and not same as oob_record conn id
            and context.connection_record
            and context.connection_record.connection_id != oob_record.connection_id
        ):
            LOGGER.debug(
                f"Oob record connection id {oob_record.connection_id} is different from"
                f" inbound message connection {context.connection_record.connection_id}",
            )
            # Mismatch in connection id's in only allowed in state await response
            # (connection id can change bc of reuse)
            if oob_record.state != OobRecord.STATE_AWAIT_RESPONSE:
                LOGGER.debug(
                    "Inbound message has incorrect connection_id "
                    f"{context.connection_record.connection_id}. Oob record "
                    f"{oob_record.oob_id} associated with connection id "
                    f"{oob_record.connection_id}"
                )
                return None

            # If the state is await response, and there are attachments we want to update
            # the connection id on the oob record. In case no request_attach is present,
            # this is handled by the reuse handlers
            if (
                oob_record.invitation.requests_attach
                and oob_record.state == OobRecord.STATE_AWAIT_RESPONSE
            ):
                LOGGER.debug(
                    f"Removing stale connection {oob_record.connection_id} due "
                    "to connection reuse"
                )
                # Remove stale connection due to connection reuse
                if oob_record.connection_id:
                    async with context.profile.session() as session:
                        old_conn_record = await ConnRecord.retrieve_by_id(
                            session, oob_record.connection_id
                        )
                        await old_conn_record.delete_record(session)

                oob_record.connection_id = context.connection_record.connection_id

        # If no attach_thread_id is stored yet we need to match the current message
        # thread_id against the attached messages in the oob invitation
        if not oob_record.attach_thread_id and oob_record.invitation.requests_attach:
            # Check if the current message thread_id corresponds to one of the invitation
            # ~thread.thid
            allowed_thread_ids = [
                self.get_thread_id(attachment.content)
                for attachment in oob_record.invitation.requests_attach
            ]

            if context.message_receipt.thread_id not in allowed_thread_ids:
                LOGGER.debug(
                    "Inbound message is for not allowed thread "
                    f"{context.message_receipt.thread_id}. Allowed "
                    f"threads are {allowed_thread_ids}"
                )
                return None

            oob_record.attach_thread_id = context.message_receipt.thread_id
        elif (
            oob_record.attach_thread_id
            and context.message_receipt.thread_id != oob_record.attach_thread_id
        ):
            LOGGER.debug(
                f"Inbound message thread id {context.message_receipt.thread_id} does not"
                f" match oob record thread id {oob_record.attach_thread_id}"
            )
            return None

        # Verify the sender key is present in their service in our record
        # If we don't have the sender verkey stored yet we can allow any key
        if oob_record.their_service and (
            context.message_receipt.recipient_verkey
            and (
                not context.message_receipt.sender_verkey
                or context.message_receipt.sender_verkey
                not in oob_record.their_service.recipient_keys
            )
        ):
            LOGGER.debug(
                "Inbound message sender verkey does not match stored service on oob"
                " record"
            )
            return None

        # If the message has a ~service decorator we save it in the oob record so we
        # can reply to this message
        if context._message._service:
            LOGGER.debug(
                "Storing service decorator in oob record %s",
                context.message._service.serialize(),
            )
            oob_record.their_service = context.message._service.serialize()

        async with context.profile.session() as session:
            # We can now remove the oob record as the connection should now be stored in
            # the exchange record itself.
            if oob_record.connection_id:
                oob_record.state = OobRecord.STATE_DONE
                await oob_record.emit_event(session)
                await oob_record.delete_record(session)
            else:
                await oob_record.save(
                    session, reason="Update their service in oob record"
                )

        return oob_record

    async def handle_message(
        self,
        profile: Profile,
        messages: List[Dict[str, Any]],
        oob_record: OobRecord,
        their_service: Optional[ServiceDecorator] = None,
    ):
        """Message handler for inbound messages."""

        supported_types = [
            CREDENTIAL_OFFER,
            CRED_20_OFFER,
            PRESENTATION_REQUEST,
            PRES_20_REQUEST,
        ]

        supported_messages = [
            message
            for message in messages
            if DIDCommPrefix.unqualify(message["@type"]) in supported_types
        ]

        if not supported_messages:
            message_str = ", ".join(supported_types)
            raise OobMessageProcessorError(
                f"None of the oob attached messages supported. Supported message types "
                f"are {message_str}"
            )

        message = supported_messages[0]
        message_str = json.dumps(message)

        async with profile.session() as session:
            message_dict, receipt = await self.wire_format.parse_message(
                session, message_str
            )

            inbound_message = InboundMessage(
                payload=message_dict,
                connection_id=oob_record.connection_id,
                receipt=receipt,
            )

            # We only need to store this data for connectionless
            # (it could be the oob record is already deleted)
            if not oob_record.connection_id:
                oob_record.attach_thread_id = self.get_thread_id(message)
                if their_service:
                    LOGGER.debug("Storing their service in oob record %s", their_service)
                    oob_record.their_service = their_service

                await oob_record.save(session)

        self._inbound_message_router(profile, inbound_message, False)

    def get_thread_id(self, message: Dict[str, Any]) -> str:
        """Extract thread id from agent message dict."""
        return message.get("~thread", {}).get("thid") or message.get("@id")



"""Presentation proposal message handler."""

from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import PresentationManager
from ..messages.presentation_problem_report import ProblemReportReason
from ..messages.presentation_proposal import PresentationProposal


class PresentationProposalHandler(BaseHandler):
    """Message handler class for presentation proposals."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for presentation proposals.

        Args:
            context: proposal context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile
        self._logger.debug("PresentationProposalHandler called with context %s", context)
        assert isinstance(context.message, PresentationProposal)
        self._logger.debug(
            "Received presentation proposal message: %s",
            context.message.serialize(as_string=True),
        )

        if not context.connection_record:
            raise HandlerException(
                "Connectionless not supported for presentation proposal"
            )
        # If connection is present it must be ready for use
        elif not context.connection_ready:
            raise HandlerException("Connection used for presentation proposal not ready")

        presentation_manager = PresentationManager(profile)
        presentation_exchange_record = await presentation_manager.receive_proposal(
            context.message, context.connection_record
        )  # mgr only creates, saves record: on exception, saving state null is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="PresentationProposalHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_respond_presentation_proposal is set, reply with proof req
        if context.settings.get("debug.auto_respond_presentation_proposal"):
            presentation_request_message = None
            try:
                (
                    presentation_exchange_record,
                    presentation_request_message,
                ) = await presentation_manager.create_bound_request(
                    presentation_exchange_record=presentation_exchange_record,
                    comment=context.message.comment,
                )
                await responder.send_reply(presentation_request_message)
            except (BaseModelError, LedgerError, StorageError) as err:
                self._logger.exception(err)
                if presentation_exchange_record:
                    async with profile.session() as session:
                        await presentation_exchange_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            presentation_exchange_record,
                            ProblemReportReason.ABANDONED.value,  # them: be vague
                        )
                    )

            trace_event(
                context.settings,
                presentation_request_message,
                outcome="PresentationProposalHandler.handle.PRESENT",
                perf_counter=r_time,
            )



from unittest import IsolatedAsyncioTestCase

from .....connections.base_manager import BaseConnectionManager
from .....messaging.responder import BaseResponder, MockResponder
from .....protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from .....protocols.did_rotate.v1_0.manager import (
    DIDRotateManager,
    ReportableDIDRotateError,
    UnrecordableKeysError,
)
from .....protocols.did_rotate.v1_0.messages.ack import RotateAck
from .....protocols.did_rotate.v1_0.messages.problem_report import RotateProblemReport
from .....protocols.did_rotate.v1_0.messages.rotate import Rotate
from .....protocols.did_rotate.v1_0.models.rotate_record import RotateRecord
from .....protocols.didcomm_prefix import DIDCommPrefix
from .....resolver.did_resolver import DIDResolver
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import message_types as test_message_types
from ..tests import MockConnRecord, test_conn_id


class TestDIDRotateManager(IsolatedAsyncioTestCase):
    test_endpoint = "http://localhost"

    async def asyncSetUp(self) -> None:
        self.responder = MockResponder()

        self.route_manager = mock.MagicMock(RouteManager)
        self.route_manager.routing_info = mock.CoroutineMock(
            return_value=([], self.test_endpoint)
        )
        self.route_manager.mediation_record_if_id = mock.CoroutineMock(return_value=None)
        self.route_manager.mediation_record_for_connection = mock.CoroutineMock(
            return_value=None
        )

        self.profile = await create_test_profile()

        self.profile.context.injector.bind_instance(BaseResponder, self.responder)
        self.profile.context.injector.bind_instance(RouteManager, self.route_manager)
        self.profile.context.injector.bind_instance(DIDResolver, DIDResolver())

        self.manager = DIDRotateManager(self.profile)
        assert self.manager.profile

    async def test_hangup(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.delete_record = mock.CoroutineMock()

        with mock.patch.object(self.responder, "send", mock.CoroutineMock()) as mock_send:
            msg = await self.manager.hangup(mock_conn_record)
            mock_conn_record.delete_record.assert_called_once()
            mock_send.assert_called_once()
            assert msg._type == DIDCommPrefix.NEW.value + "/" + test_message_types.HANGUP

    async def test_receive_hangup(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.delete_record = mock.CoroutineMock()

        await self.manager.receive_hangup(mock_conn_record)
        mock_conn_record.delete_record.assert_called_once()

    async def test_rotate_my_did(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        test_to_did = "did:peer:2:testdid"

        with mock.patch.object(self.responder, "send", mock.CoroutineMock()) as mock_send:
            msg = await self.manager.rotate_my_did(mock_conn_record, test_to_did)
            mock_send.assert_called_once()
            assert msg._type == DIDCommPrefix.NEW.value + "/" + test_message_types.ROTATE

    async def test_receive_rotate(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)

        test_to_did = "did:peer:2:testdid"

        record = await self.manager.receive_rotate(
            mock_conn_record, Rotate(to_did=test_to_did)
        )

        assert record.RECORD_TYPE == RotateRecord.RECORD_TYPE
        assert record.role == record.ROLE_OBSERVING
        assert record.state == record.STATE_ROTATE_RECEIVED
        assert record.connection_id == mock_conn_record.connection_id

    async def test_receive_rotate_x(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)

        test_to_did = "did:badmethod:1:testdid"
        test_problem_report = ReportableDIDRotateError(
            RotateProblemReport(problem_items=[{"did": test_to_did}])
        )

        with (
            mock.patch.object(
                self.manager, "ensure_supported_did", side_effect=test_problem_report
            ),
            mock.patch.object(self.responder, "send", mock.CoroutineMock()) as mock_send,
        ):
            await self.manager.receive_rotate(
                mock_conn_record, Rotate(to_did=test_to_did)
            )
            mock_send.assert_called_once_with(
                test_problem_report.message,
                connection_id=mock_conn_record.connection_id,
            )

    @mock.patch.object(
        BaseConnectionManager,
        "record_keys_for_resolvable_did",
        mock.CoroutineMock(),
    )
    async def test_commit_rotate(self, *_):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.save = mock.CoroutineMock()

        test_to_did = "did:peer:2:testdid"

        record = await self.manager.receive_rotate(
            mock_conn_record, Rotate(to_did=test_to_did)
        )
        await self.manager.commit_rotate(mock_conn_record, record)

        assert record.state == RotateRecord.STATE_ACK_SENT

    @mock.patch.object(
        BaseConnectionManager,
        "record_keys_for_resolvable_did",
        mock.CoroutineMock(),
    )
    async def test_commit_rotate_x_no_new_did(self, *_):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.save = mock.CoroutineMock()

        test_to_did = "did:peer:2:testdid"

        record = await self.manager.receive_rotate(
            mock_conn_record, Rotate(to_did=test_to_did)
        )
        record.new_did = None

        with self.assertRaises(ValueError):
            await self.manager.commit_rotate(mock_conn_record, record)

    async def test_commit_rotate_x_unrecordable_keys(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.save = mock.CoroutineMock()

        test_to_did = "did:peer:2:testdid"

        record = await self.manager.receive_rotate(
            mock_conn_record, Rotate(to_did=test_to_did)
        )

        with self.assertRaises(UnrecordableKeysError):
            await self.manager.commit_rotate(mock_conn_record, record)

    @mock.patch.object(
        BaseConnectionManager,
        "clear_connection_targets_cache",
        mock.CoroutineMock(),
    )
    async def test_receive_ack(self, *_):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.save = mock.CoroutineMock()

        with mock.patch.object(
            RotateRecord,
            "retrieve_by_thread_id",
            return_value=mock.CoroutineMock(
                return_value=mock.MagicMock(
                    new_did="did:peer:2:testdid", delete_record=mock.CoroutineMock()
                )
            ),
        ) as mock_rotate_record:
            await self.manager.receive_ack(mock_conn_record, RotateAck())

            mock_conn_record.save.assert_called_once()
            mock_rotate_record.return_value.delete_record.assert_called_once()

    async def test_receive_ack_x(self):
        mock_conn_record = MockConnRecord(test_conn_id, True)
        mock_conn_record.save = mock.CoroutineMock()

        with mock.patch.object(
            RotateRecord,
            "retrieve_by_thread_id",
            return_value=mock.CoroutineMock(),
        ) as mock_rotate_record:
            mock_rotate_record.return_value.new_did = None
            with self.assertRaises(ValueError):
                await self.manager.receive_ack(mock_conn_record, RotateAck())

    async def test_receive_problem_report(self):
        test_to_did = "did:badmethod:1:testdid"
        mock_problem_report = RotateProblemReport(
            description={"code": 123},
            problem_items=[{"did": test_to_did}],
        )

        with mock.patch.object(
            RotateRecord,
            "retrieve_by_thread_id",
            return_value=mock.CoroutineMock(
                return_value=mock.MagicMock(save=mock.CoroutineMock())
            ),
        ) as mock_rotate_record:
            await self.manager.receive_problem_report(mock_problem_report)

            mock_rotate_record.return_value.save.assert_called_once()



"""Askar storage implementation of VC holder interface."""

import json
from typing import Mapping, Optional, Sequence

from ...askar.profile import AskarProfile
from ..askar import AskarStorage, AskarStorageSearch, AskarStorageSearchSession
from ..record import StorageRecord
from .base import VCHolder, VCRecordSearch
from .vc_record import VCRecord
from .xform import VC_CRED_RECORD_TYPE


class AskarVCHolder(VCHolder):
    """Askar VC record storage class."""

    def __init__(self, profile: AskarProfile):
        """Initialize the Indy-SDK VC holder instance."""
        self._profile = profile

    def build_type_or_schema_query(self, uri_list: Sequence[str]) -> dict:
        """Build and return indy-specific type_or_schema_query."""
        type_or_schema_query = {}
        for uri in uri_list:
            q = {"$or": [{"type": uri}, {"schema": uri}]}
            if type_or_schema_query:
                if "$and" not in type_or_schema_query:
                    type_or_schema_query = {"$and": [type_or_schema_query]}
                type_or_schema_query["$and"].append(q)
            else:
                type_or_schema_query = q
        return type_or_schema_query

    async def store_credential(self, cred: VCRecord):
        """Add a new VC record to the store.

        Args:
            cred: The VCRecord instance to store
        Raises:
            StorageDuplicateError: If the record_id is not unique

        """
        record = vc_to_storage_record(cred)
        async with self._profile.session() as session:
            await AskarStorage(session).add_record(record)

    async def retrieve_credential_by_id(self, record_id: str) -> VCRecord:
        """Fetch a VC record by its record ID.

        Raises:
            StorageNotFoundError: If the record is not found

        """
        async with self._profile.session() as session:
            record = await AskarStorage(session).get_record(
                VC_CRED_RECORD_TYPE, record_id
            )
        return storage_to_vc_record(record)

    async def retrieve_credential_by_given_id(self, given_id: str) -> VCRecord:
        """Fetch a VC record by its given ID ('id' property).

        Raises:
            StorageNotFoundError: If the record is not found

        """
        async with self._profile.session() as session:
            record = await AskarStorage(session).find_record(
                VC_CRED_RECORD_TYPE, {"given_id": given_id}
            )
        return storage_to_vc_record(record)

    async def delete_credential(self, cred: VCRecord):
        """Remove a previously-stored VC record.

        Raises:
            StorageNotFoundError: If the record is not found

        """
        async with self._profile.session() as session:
            await AskarStorage(session).delete_record(vc_to_storage_record(cred))

    def search_credentials(
        self,
        contexts: Sequence[str] = None,
        types: Sequence[str] = None,
        schema_ids: Sequence[str] = None,
        issuer_id: Optional[str] = None,
        subject_ids: Optional[str] = None,
        proof_types: Sequence[str] = None,
        given_id: Optional[str] = None,
        tag_query: Optional[Mapping] = None,
        pd_uri_list: Sequence[str] = None,
    ) -> "VCRecordSearch":
        """Start a new VC record search.

        Args:
            contexts: An inclusive list of JSON-LD contexts to match
            types: An inclusive list of JSON-LD types to match
            schema_ids: An inclusive list of credential schema identifiers
            issuer_id: The ID of the credential issuer
            subject_ids: The IDs of credential subjects all of which to match
            proof_types: The signature suite types used for the proof objects.
            given_id: The given id of the credential
            tag_query: A tag filter clause
            pd_uri_list: A list of presentation definition URIs to match

        """

        def _match_any(query: list, k, vals):
            if vals is None:
                pass
            elif len(vals) > 1:
                query.append({"$or": [{k: v for v in vals}]})
            else:
                query.append({k: vals[0]})

        def _make_custom_query(query):
            result = {}
            for k, v in query.items():
                if isinstance(v, (list, set)) and k != "$exist":
                    result[k] = [_make_custom_query(cl) for cl in v]
                elif k.startswith("$"):
                    result[k] = v
                else:
                    result[f"cstm:{k}"] = v
            return result

        query = []
        _match_any(query, "context", contexts)
        _match_any(query, "type", types)
        _match_any(query, "schema", schema_ids)
        _match_any(query, "subject", subject_ids)
        _match_any(query, "proof_type", proof_types)
        if issuer_id:
            query.append({"issuer_id": issuer_id})
        if given_id:
            query.append({"given_id": given_id})
        if tag_query:
            query.append(_make_custom_query(tag_query))
        if pd_uri_list:
            query.append(self.build_type_or_schema_query(pd_uri_list))
        query = {"$and": query} if query else None
        search = AskarStorageSearch(self._profile).search_records(
            VC_CRED_RECORD_TYPE, query
        )
        return AskarVCRecordSearch(search)


class AskarVCRecordSearch(VCRecordSearch):
    """Askar storage search for VC records."""

    def __init__(self, search: AskarStorageSearchSession):
        """Initialize the Askar VC record search."""
        self._search = search

    async def close(self):
        """Dispose of the search query."""
        await self._search.close()

    async def fetch(self, max_count: Optional[int] = None) -> Sequence[VCRecord]:
        """Fetch the next list of VC records from the store.

        Args:
            max_count: Max number of records to return. If not provided,
              defaults to the backend's preferred page size

        Returns:
            A list of `VCRecord` instances

        """
        rows = await self._search.fetch(max_count)
        records = [storage_to_vc_record(r) for r in rows]
        return records


def storage_to_vc_record(record: StorageRecord) -> VCRecord:
    """Convert an Askar stored record into a VC record."""

    def _make_set(val) -> set:
        if isinstance(val, str):
            return {val}
        else:
            return set(val)

    cred_tags = {}
    contexts = set()
    types = set()
    schema_ids = set()
    subject_ids = set()
    proof_types = set()
    issuer_id = None
    given_id = None
    for tagname, tagval in (record.tags or {}).items():
        if tagname == "context":
            contexts = _make_set(tagval)
        elif tagname == "type":
            types = _make_set(tagval)
        elif tagname == "schema":
            schema_ids = _make_set(tagval)
        elif tagname == "subject":
            subject_ids = _make_set(tagval)
        elif tagname == "proof_type":
            proof_types = _make_set(tagval)
        elif tagname == "issuer_id":
            issuer_id = tagval
        elif tagname == "given_id":
            given_id = tagval
        elif tagname.startswith("cstm:"):
            cred_tags[tagname[5:]] = tagval
    return VCRecord(
        contexts=contexts,
        expanded_types=types,
        schema_ids=schema_ids,
        issuer_id=issuer_id,
        subject_ids=subject_ids,
        proof_types=proof_types,
        cred_value=json.loads(record.value),
        given_id=given_id,
        cred_tags=cred_tags,
        record_id=record.id,
    )


def vc_to_storage_record(cred: VCRecord) -> StorageRecord:
    """Convert a VC record into an Askar stored record."""
    tags = {}
    tags["context"] = set(cred.contexts)
    tags["type"] = set(cred.expanded_types)
    tags["schema"] = set(cred.schema_ids)
    tags["subject"] = set(cred.subject_ids)
    tags["proof_type"] = set(cred.proof_types)
    if cred.issuer_id:
        tags["issuer_id"] = cred.issuer_id
    if cred.given_id:
        tags["given_id"] = cred.given_id
    for tagname, tagval in (cred.cred_tags or {}).items():
        tags[f"cstm:{tagname}"] = tagval

    return StorageRecord(
        VC_CRED_RECORD_TYPE,
        json.dumps(cred.cred_value),
        tags,
        cred.record_id,
    )



import json
from copy import deepcopy
from time import time
from unittest import TestCase

import pytest

from ....ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ....messaging.util import canon
from ....multitenant.base import BaseMultitenantManager
from ....multitenant.manager import MultitenantManager
from ....protocols.didcomm_prefix import DIDCommPrefix
from ....tests import mock
from ....utils.testing import create_test_profile
from ..non_rev_interval import IndyNonRevocationInterval
from ..predicate import Predicate
from ..pres_preview import (
    PRESENTATION_PREVIEW,
    IndyPresAttrSpec,
    IndyPresPredSpec,
    IndyPresPreview,
)

S_ID = {
    "score": "NcYxiDXkpYi6ov5FcYDi1e:2:score:1.0",
    "membership": "NcYxiDXkpYi6ov5FcYDi1e:2:membership:1.0",
}
CD_ID = {name: f"NcYxiDXkpYi6ov5FcYDi1e:3:CL:{S_ID[name]}:tag1" for name in S_ID}
PRES_PREVIEW = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(
            name="player", cred_def_id=CD_ID["score"], value="Richie Knucklez"
        ),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID["score"],
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
        ),
    ],
    predicates=[
        IndyPresPredSpec(
            name="highScore",
            cred_def_id=CD_ID["score"],
            predicate=">=",
            threshold=1000000,
        )
    ],
)
PRES_PREVIEW_ATTR_NAMES = IndyPresPreview(
    attributes=[
        IndyPresAttrSpec(
            name="player",
            cred_def_id=CD_ID["score"],
            value="Richie Knucklez",
            referent="reft-0",
        ),
        IndyPresAttrSpec(
            name="screenCapture",
            cred_def_id=CD_ID["score"],
            mime_type="image/png",
            value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
            referent="reft-0",
        ),
        IndyPresAttrSpec(
            name="member",
            cred_def_id=CD_ID["membership"],
            value="Richard Hand",
            referent="reft-1",
        ),
        IndyPresAttrSpec(
            name="since",
            cred_def_id=CD_ID["membership"],
            value="2020-01-01",
            referent="reft-1",
        ),
    ]
)
INDY_PROOF_REQ = json.loads(
    f"""{{
    "name": "proof-req",
    "version": "1.0",
    "nonce": "12345",
    "requested_attributes": {{
        "0_player_uuid": {{
            "name": "player",
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID["score"]}"
                }}
            ]
        }},
        "1_screencapture_uuid": {{
            "name": "screenCapture",
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID["score"]}"
                }}
            ]
        }}
    }},
    "requested_predicates": {{
        "0_highscore_GE_uuid": {{
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID["score"]}"
                }}
            ]
        }}
    }}
}}"""
)
INDY_PROOF_REQ_ATTR_NAMES = json.loads(
    f"""{{
    "name": "proof-req",
    "version": "1.0",
    "nonce": "12345",
    "requested_attributes": {{
        "0_player_uuid": {{
            "names": ["player", "screenCapture"],
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID["score"]}"
                }}
            ]
        }},
        "1_member_uuid": {{
            "names": ["member", "since"],
            "restrictions": [
                {{
                    "cred_def_id": "{CD_ID["membership"]}"
                }}
            ]
        }}
    }},
    "requested_predicates": {{}}
}}"""
)


class TestIndyPresAttrSpec(TestCase):
    """Presentation-preview attribute specification tests"""

    def test_posture(self):
        self_attested = IndyPresAttrSpec(name="ident", cred_def_id=None, value="655321")
        assert self_attested.posture == IndyPresAttrSpec.Posture.SELF_ATTESTED

        revealed = IndyPresAttrSpec(
            name="ident", cred_def_id=CD_ID["score"], value="655321"
        )
        assert revealed.posture == IndyPresAttrSpec.Posture.REVEALED_CLAIM

        unrevealed = IndyPresAttrSpec(name="ident", cred_def_id=CD_ID["score"])
        assert unrevealed.posture == IndyPresAttrSpec.Posture.UNREVEALED_CLAIM

        no_posture = IndyPresAttrSpec(name="no_spec")
        assert no_posture.posture is None

    def test_list_plain(self):
        by_list = IndyPresAttrSpec.list_plain(
            plain={"ident": "655321", " Given Name ": "Alexander DeLarge"},
            cred_def_id=CD_ID["score"],
        )
        explicit = [
            IndyPresAttrSpec(name="ident", cred_def_id=CD_ID["score"], value="655321"),
            IndyPresAttrSpec(
                name="givenname", cred_def_id=CD_ID["score"], value="Alexander DeLarge"
            ),
        ]

        # order could be askew
        for listp in by_list:
            assert any(xp == listp for xp in explicit)
        assert len(explicit) == len(by_list)

    def test_list_plain_share_referent(self):
        by_list = IndyPresAttrSpec.list_plain(
            plain={"ident": "655321", " Given Name ": "Alexander DeLarge"},
            cred_def_id=CD_ID["score"],
            referent="dummy",
        )
        explicit = [
            IndyPresAttrSpec(
                name="ident",
                cred_def_id=CD_ID["score"],
                value="655321",
                referent="dummy",
            ),
            IndyPresAttrSpec(
                name="givenname",
                cred_def_id=CD_ID["score"],
                value="Alexander DeLarge",
                referent="dummy",
            ),
        ]

        # order could be askew
        for listp in by_list:
            assert any(xp == listp for xp in explicit)
        assert len(explicit) == len(by_list)

    def test_eq(self):
        attr_specs_none_plain = [
            IndyPresAttrSpec(name="name", value="value"),
            IndyPresAttrSpec(name="name", value="value", mime_type=None),
            IndyPresAttrSpec(name=" NAME ", value="value"),
        ]
        attr_specs_different = [
            IndyPresAttrSpec(name="name", value="dmFsdWU=", mime_type="image/png"),
            IndyPresAttrSpec(name="name", value="value", cred_def_id="cred_def_id"),
            IndyPresAttrSpec(name="name", value="distinct value", mime_type=None),
            IndyPresAttrSpec(name="distinct name", value="value", mime_type=None),
            IndyPresAttrSpec(name="name", value="dmFsdWU=", mime_type=None),
            IndyPresAttrSpec(name="name"),
            IndyPresAttrSpec(
                name="name", value="value", cred_def_id="cred_def_id", referent="reft-0"
            ),
            IndyPresAttrSpec(
                name="name", value="value", cred_def_id="cred_def_id", referent="reft-1"
            ),
        ]

        for lhs in attr_specs_none_plain:
            for rhs in attr_specs_different:
                assert lhs != rhs

        for lidx in range(len(attr_specs_none_plain) - 1):
            for ridx in range(lidx + 1, len(attr_specs_none_plain)):
                assert attr_specs_none_plain[lidx] == attr_specs_none_plain[ridx]

        for lidx in range(len(attr_specs_different) - 1):
            for ridx in range(lidx + 1, len(attr_specs_different)):
                assert attr_specs_different[lidx] != attr_specs_different[ridx]

    def test_deserialize(self):
        """Test deserialization."""
        dump = json.dumps(
            {
                "name": "PLAYER",
                "cred_def_id": CD_ID["score"],
                "value": "Richie Knucklez",
            }
        )

        attr_spec = IndyPresAttrSpec.deserialize(dump)
        assert type(attr_spec) is IndyPresAttrSpec
        assert canon(attr_spec.name) == "player"

        dump = json.dumps(
            {
                "name": "PLAYER",
                "cred_def_id": CD_ID["score"],
                "value": "Richie Knucklez",
                "referent": "0",
            }
        )

        attr_spec = IndyPresAttrSpec.deserialize(dump)
        assert type(attr_spec) is IndyPresAttrSpec
        assert canon(attr_spec.name) == "player"

    def test_serialize(self):
        """Test serialization."""

        attr_spec_dict = PRES_PREVIEW.attributes[0].serialize()
        assert attr_spec_dict == {
            "name": "player",
            "cred_def_id": CD_ID["score"],
            "value": "Richie Knucklez",
        }

        attr_spec_dict = PRES_PREVIEW_ATTR_NAMES.attributes[0].serialize()
        assert attr_spec_dict == {
            "name": "player",
            "cred_def_id": CD_ID["score"],
            "value": "Richie Knucklez",
            "referent": "reft-0",
        }


class TestIndyPresPredSpec(TestCase):
    """Presentation predicate specification tests"""

    def test_deserialize(self):
        """Test deserialization."""
        dump = json.dumps(
            {
                "name": "HIGH SCORE",
                "cred_def_id": CD_ID["score"],
                "predicate": ">=",
                "threshold": 1000000,
            }
        )

        pred_spec = IndyPresPredSpec.deserialize(dump)
        assert type(pred_spec) is IndyPresPredSpec
        assert canon(pred_spec.name) == "highscore"

    def test_serialize(self):
        """Test serialization."""

        pred_spec_dict = PRES_PREVIEW.predicates[0].serialize()
        assert pred_spec_dict == {
            "name": "highScore",
            "cred_def_id": CD_ID["score"],
            "predicate": ">=",
            "threshold": 1000000,
        }

    def test_eq(self):
        """Test equality operator."""

        pred_spec_a = IndyPresPredSpec(
            name="a",
            cred_def_id=CD_ID["score"],
            predicate=Predicate.GE.value.math,
            threshold=0,
        )
        pred_spec_b = IndyPresPredSpec(
            name="b",
            cred_def_id=CD_ID["score"],
            predicate=Predicate.GE.value.math,
            threshold=0,
        )
        assert pred_spec_a != pred_spec_b

        pred_spec_a.name = "b"
        assert pred_spec_a == pred_spec_b

        pred_spec_a.predicate = Predicate.LE.value.math
        assert pred_spec_a != pred_spec_b

        pred_spec_a.predicate = Predicate.GE.value.math
        assert pred_spec_a == pred_spec_b

        pred_spec_a.threshold = 100
        assert pred_spec_a != pred_spec_b

        pred_spec_a.threshold = 0
        pred_spec_a.cred_def_id = None
        assert pred_spec_a != pred_spec_b


class TestIndyPresPreviewAsync:
    """Presentation preview tests"""

    @pytest.mark.asyncio
    async def test_to_indy_proof_request(self):
        """Test presentation preview to indy proof request."""

        indy_proof_req = await PRES_PREVIEW.indy_proof_request(
            **{k: INDY_PROOF_REQ[k] for k in ("name", "version", "nonce")}
        )

        assert indy_proof_req == INDY_PROOF_REQ

    @pytest.mark.asyncio
    async def test_to_indy_proof_request_attr_names(self):
        """Test presentation preview to indy proof request."""

        indy_proof_req = await PRES_PREVIEW_ATTR_NAMES.indy_proof_request(
            **{k: INDY_PROOF_REQ_ATTR_NAMES[k] for k in ("name", "version", "nonce")}
        )

        assert indy_proof_req == INDY_PROOF_REQ_ATTR_NAMES

    @pytest.mark.asyncio
    async def test_to_indy_proof_request_self_attested(self):
        """Test presentation preview to indy proof request with self-attested values."""

        pres_preview_selfie = deepcopy(PRES_PREVIEW)
        for attr_spec in pres_preview_selfie.attributes:
            attr_spec.cred_def_id = None

        indy_proof_req_selfie = await pres_preview_selfie.indy_proof_request(
            **{k: INDY_PROOF_REQ[k] for k in ("name", "version", "nonce")}
        )

        assert not any(
            "restrictions" in attr_spec
            for attr_spec in indy_proof_req_selfie["requested_attributes"].values()
        )

    @pytest.mark.asyncio
    async def test_to_indy_proof_request_revo_default_interval(self):
        """Test pres preview to indy proof req with revocation support, defaults."""

        copy_indy_proof_req = deepcopy(INDY_PROOF_REQ)

        pres_preview = deepcopy(PRES_PREVIEW)
        self.profile = await create_test_profile()
        context = self.profile.context
        context.injector.bind_instance(
            IndyLedgerRequestsExecutor, IndyLedgerRequestsExecutor(self.profile)
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor, "get_ledger_for_identifier"
        ) as mock_get_ledger:
            mock_get_ledger.return_value = (
                None,
                mock.MagicMock(
                    get_credential_definition=mock.CoroutineMock(
                        return_value={"value": {"revocation": {"...": "..."}}}
                    )
                ),
            )
            indy_proof_req_revo = await pres_preview.indy_proof_request(
                **{k: INDY_PROOF_REQ[k] for k in ("name", "version", "nonce")},
                profile=self.profile,
            )

        for uuid, attr_spec in indy_proof_req_revo["requested_attributes"].items():
            assert set(attr_spec.get("non_revoked", {}).keys()) == {"from", "to"}
            copy_indy_proof_req["requested_attributes"][uuid]["non_revoked"] = attr_spec[
                "non_revoked"
            ]
        for uuid, pred_spec in indy_proof_req_revo["requested_predicates"].items():
            assert set(pred_spec.get("non_revoked", {}).keys()) == {"from", "to"}
            copy_indy_proof_req["requested_predicates"][uuid]["non_revoked"] = pred_spec[
                "non_revoked"
            ]

        assert copy_indy_proof_req == indy_proof_req_revo

    @pytest.mark.asyncio
    async def test_to_indy_proof_request_revo(self):
        """Test pres preview to indy proof req with revocation support, interval."""

        EPOCH_NOW = int(time())
        copy_indy_proof_req = deepcopy(INDY_PROOF_REQ)

        pres_preview = deepcopy(PRES_PREVIEW)
        self.profile = await create_test_profile()
        self.profile.settings["ledger.ledger_config_list"] = [{"id": "test"}]
        context = self.profile.context
        context.injector.bind_instance(
            IndyLedgerRequestsExecutor, IndyLedgerRequestsExecutor(self.profile)
        )
        context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor, "get_ledger_for_identifier"
        ) as mock_get_ledger:
            mock_get_ledger.return_value = (
                None,
                mock.MagicMock(
                    get_credential_definition=mock.CoroutineMock(
                        return_value={"value": {"revocation": {"...": "..."}}}
                    )
                ),
            )
            indy_proof_req_revo = await pres_preview.indy_proof_request(
                **{k: INDY_PROOF_REQ[k] for k in ("name", "version", "nonce")},
                profile=self.profile,
                non_revoc_intervals={
                    CD_ID[s_id]: IndyNonRevocationInterval(1234567890, EPOCH_NOW)
                    for s_id in S_ID
                },
            )

        for uuid, attr_spec in indy_proof_req_revo["requested_attributes"].items():
            assert set(attr_spec.get("non_revoked", {}).keys()) == {"from", "to"}
            copy_indy_proof_req["requested_attributes"][uuid]["non_revoked"] = attr_spec[
                "non_revoked"
            ]
        for uuid, pred_spec in indy_proof_req_revo["requested_predicates"].items():
            assert set(pred_spec.get("non_revoked", {}).keys()) == {"from", "to"}
            copy_indy_proof_req["requested_predicates"][uuid]["non_revoked"] = pred_spec[
                "non_revoked"
            ]

        assert copy_indy_proof_req == indy_proof_req_revo

    @pytest.mark.asyncio
    async def test_satisfaction(self):
        """Test presentation preview predicate satisfaction."""

        pred_spec = IndyPresPredSpec(
            name="highScore",
            cred_def_id=CD_ID["score"],
            predicate=Predicate.GE.value.math,
            threshold=1000000,
        )
        attr_spec = IndyPresAttrSpec(
            name="HIGHSCORE", cred_def_id=CD_ID["score"], value=1234567
        )
        assert attr_spec.satisfies(pred_spec)

        attr_spec = IndyPresAttrSpec(
            name="HIGHSCORE", cred_def_id=CD_ID["score"], value=985260
        )
        assert not attr_spec.satisfies(pred_spec)


class TestIndyPresPreview(TestCase):
    """Presentation preview tests"""

    def test_init(self):
        """Test initializer."""
        assert PRES_PREVIEW.attributes
        assert PRES_PREVIEW.predicates
        assert PRES_PREVIEW.has_attr_spec(
            cred_def_id=CD_ID["score"], name="player", value="Richie Knucklez"
        )

    def test_type(self):
        """Test type."""
        assert PRES_PREVIEW._type == DIDCommPrefix.qualify_current(PRESENTATION_PREVIEW)

    def test_deserialize(self):
        """Test deserialization."""
        dump = {
            "@type": DIDCommPrefix.qualify_current(PRESENTATION_PREVIEW),
            "attributes": [
                {
                    "name": "player",
                    "cred_def_id": CD_ID["score"],
                    "value": "Richie Knucklez",
                },
                {
                    "name": "screenCapture",
                    "cred_def_id": CD_ID["score"],
                    "mime-type": "image/png",
                    "value": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                },
            ],
            "predicates": [
                {
                    "name": "highScore",
                    "cred_def_id": CD_ID["score"],
                    "predicate": ">=",
                    "threshold": 1000000,
                }
            ],
        }

        preview = IndyPresPreview.deserialize(dump)
        assert type(preview) is IndyPresPreview

    def test_serialize(self):
        """Test serialization."""

        preview_dict = PRES_PREVIEW.serialize()
        assert preview_dict == {
            "@type": DIDCommPrefix.qualify_current(PRESENTATION_PREVIEW),
            "attributes": [
                {
                    "name": "player",
                    "cred_def_id": CD_ID["score"],
                    "value": "Richie Knucklez",
                },
                {
                    "name": "screenCapture",
                    "cred_def_id": CD_ID["score"],
                    "mime-type": "image/png",
                    "value": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                },
            ],
            "predicates": [
                {
                    "name": "highScore",
                    "cred_def_id": CD_ID["score"],
                    "predicate": ">=",
                    "threshold": 1000000,
                }
            ],
        }

    def test_eq(self):
        pres_preview_a = IndyPresPreview.deserialize(PRES_PREVIEW.serialize())
        assert pres_preview_a == PRES_PREVIEW

        pres_preview_a.predicates = []
        assert pres_preview_a != PRES_PREVIEW
